import typing
from io import BytesIO
from pathlib import Path
from typing import List, Optional, Union

import solara
from solara.alias import rw

if typing.TYPE_CHECKING:
    import numpy as np
    import PIL.Image


@solara.component
def Image(
    image: Union[str, Path, "np.ndarray", bytes, "PIL.Image.Image"],
    width: Optional[str] = None,
    format="png",
    classes: List[str] = [],
):
    """Displays an image from a URL, Path, numpy data, bytes or PIL image.

    If passed as bytes, the image is assumed to be in the format specified by the `format` argument.

    ## Examples

    ### Displaying an image from a Path

    ```solara
    import solara
    import solara.website
    from pathlib import Path


    image_path = Path(solara.website.__file__).parent / "public/beach.jpeg"

    @solara.component
    def Page():
        solara.Image(image_path)
    ```

    ### Displaying an image from a URL

    ```solara
    import solara


    image_url = "/static/public/beach.jpeg"

    @solara.component
    def Page():
        solara.Image(image_url)
        display(image_url)

    ```

    ### Displaying an image from a numpy array

    ```solara

    import solara
    import solara.website
    from pathlib import Path
    import numpy as np
    import PIL.Image


    image_path = Path(solara.website.__file__).parent / "public/beach.jpeg"
    image_ndarray = np.asarray(PIL.Image.open(image_path))

    @solara.component
    def Page():
        solara.Image(image_ndarray)
        display(image_ndarray)
    ```

    ### Displaying an image from bytes

    ```solara

    import solara
    import solara.website
    from pathlib import Path
    import PIL.Image


    image_path = Path(solara.website.__file__).parent / "public/beach.jpeg"
    image_bytes = image_path.read_bytes()

    @solara.component
    def Page():
        solara.Image(image_bytes, format="jpeg")
        display(image_bytes[:100])
    ```

    ### Displaying an image from a PIL Image

    ```solara

    import solara
    import solara.website
    from pathlib import Path
    import PIL.Image


    image_path = Path(solara.website.__file__).parent / "public/beach.jpeg"
    image = PIL.Image.open(image_path)

    @solara.component
    def Page():
        solara.Image(image)
    ```


    ## Arguments
     * `image`: URL, Path, numpy data, bytes or PIL Image.
     * `width`: Width of the image, by default (None) the image is displayed at its original size.
        Other options are: "100%" (to full its parent size), "100px" (to display it at 100 pixels wide),
        or any other CSS width value.
     * `format`: Format of the image, only used when image is of type bytes, or a PIL Image is passed.
     * `classes`: List of CSS classes to add to the image.
    """
    layout = {}
    if width:
        layout["width"] = width
    classes = ["solara-image"] + classes
    if isinstance(image, (str, Path)):
        path = Path(image)
        if path.exists():
            import ipywidgets

            format = ipywidgets.Image._guess_format("image", str(path))
            value = ipywidgets.Image._load_file_value(str(path))
            return rw.Image(
                value=value,
                format=format,
                _dom_classes=classes,  # type: ignore
                layout=layout,
            )
        elif isinstance(image, str):
            url_data = image.encode("utf8")
            return rw.Image(
                value=url_data,
                format="url",
                _dom_classes=classes,  # type: ignore
                layout=layout,
            )
    elif isinstance(image, bytes):
        return rw.Image(
            value=image,
            format=format,
            _dom_classes=classes,  # type: ignore
            layout=layout,
        )
    elif solara.util.isinstanceof(image, "PIL.Image:Image"):
        f = BytesIO()
        image.save(f, format=format)  # type: ignore
        value = f.getvalue()
        return rw.Image(
            value=value,
            format="png",
            _dom_classes=classes,  # type: ignore
            layout=layout,
        )
    elif solara.util.isinstanceof(image, "numpy:ndarray"):
        value = solara.util.numpy_to_image(image, format="png")
        return rw.Image(
            value=value,
            format="png",
            _dom_classes=classes,  # type: ignore
            layout=layout,
        )
    raise TypeError(f"Only support URL, path or numpy array, not {image}")



import contextlib
import logging
import os
import pdb
import traceback

from rich import print

logger = logging.getLogger("solara.server")


def start_error(title, msg, exception: Exception = None):
    if exception:
        traceback.print_exception(None, exception, exception.__traceback__)
    print(f"[red]{title}:\n\t[blue]{msg}")  # noqa
    os._exit(-1)


@contextlib.contextmanager
def pdb_guard():
    from . import settings

    try:
        yield
    except Exception:
        if settings.main.use_pdb:
            logger.exception("Exception, will be handled by debugger")
            pdb.post_mortem()
        else:
            raise



from typing import Callable, Dict, List, Optional, TypeVar, Union, cast, overload

import ipyvuetify as v
import reacton
from typing_extensions import Literal

import solara
from solara.alias import rv
from solara.util import _combine_classes

T = TypeVar("T")


def _get_button_value(button: reacton.core.Element):
    if "value" in button.kwargs:
        value = button.kwargs["value"]
    else:
        value = button.kwargs.get("label")
        if value is None and button.args:
            value = button.args[0]
    return value


@overload
@solara.value_component(None)
def ToggleButtonsSingle(
    value: Union[T, solara.Reactive[T]],
    values: List[T] = ...,
    children: List[reacton.core.Element] = ...,
    on_value: Optional[Callable[[T], None]] = ...,
    dense: bool = ...,
    mandatory: Literal[True] = ...,
    classes: List[str] = ...,
    style: Union[str, Dict[str, str], None] = ...,
) -> reacton.core.ValueElement[v.BtnToggle, T]:
    ...


@overload
@solara.value_component(None)
def ToggleButtonsSingle(
    value: Union[Optional[T], solara.Reactive[Optional[T]]] = None,
    values: List[T] = ...,
    children: List[reacton.core.Element] = ...,
    on_value: Optional[Callable[[Optional[T]], None]] = ...,
    dense: bool = ...,
    mandatory: Literal[False] = ...,
    classes: List[str] = ...,
    style: Union[str, Dict[str, str], None] = ...,
) -> reacton.core.ValueElement[v.BtnToggle, T]:
    ...


@solara.value_component(None)
def ToggleButtonsSingle(
    value: Union[None, T, Optional[T], solara.Reactive[T], solara.Reactive[Optional[T]]] = None,
    values: List[T] = [],
    children: List[reacton.core.Element] = [],
    on_value: Optional[Callable[[T], None]] = None,
    dense: bool = False,
    mandatory: bool = True,
    classes: List[str] = [],
    style: Union[str, Dict[str, str], None] = None,
) -> reacton.core.ValueElement[v.BtnToggle, T]:
    """ToggleButtons for selecting a single value.

    ```solara
    import solara

    foods = ["Kiwi", "Banana", "Apple"]
    food = solara.reactive("Banana")


    @solara.component
    def Page():
        with solara.Card("My favorite food"):
            solara.ToggleButtonsSingle(value=food, values=foods)
            solara.Markdown(f"**Selected**: {food.value}")
    ```

    ### Using Buttons

    Instead of passing the values as a list, we can also use the buttons as children.
    This allows us to customize the buttons, for example to add an icon.

    ```solara
    import solara

    direction = solara.reactive("left")


    @solara.component
    def Page():
        with solara.Card("Pick a direction"):
            # instead of using the values argument, we can use the buttons as children
            # the label of the button will be used as value, if no value is given.
            with solara.ToggleButtonsSingle(value=direction):
                # note that the label and the value are different
                solara.Button("Up", icon_name="mdi-arrow-up-bold", value="up", text=True)
                solara.Button("Down", icon_name="mdi-arrow-down-bold", value="down", text=True)
                solara.Button("Left", icon_name="mdi-arrow-left-bold", value="left", text=True)
                solara.Button("Right", icon_name="mdi-arrow-right-bold", value="right", text=True)
            solara.Markdown(f"**Selected**: {direction.value}")
    ```

    ## Arguments

     * `value`: The currently selected value.
     * `values`: List of values to select from.
     * `children`: List of buttons to use as values.
     * `on_value`: Callback to call when the value changes.
     * `dense`: Whether to use a dense (smaller) style.
     * `mandatory`: Whether a choice is mandatory.
     * `style`: CSS style to apply to the top level element.
     * `classes`: List of CSS classes to be applied to the top level element.
    """
    class_ = _combine_classes(classes)
    style_flat = solara.util._flatten_style(style)
    # TODO: make type safe
    # typing is ignored below due to an issue with the typing; The combination of value being T and on_value being of type Callback[[Optional[T]], None] is
    # not allowed to be passed to use_reactive. We also do not allow this by using our overloads, but this information seems lost at this point by
    # the typechecker
    reactive_value = solara.use_reactive(value, on_value)  # type: ignore
    children = [solara.Button(label=str(value)) for value in values] + children
    values = values + [_get_button_value(button) for button in children]  # type: ignore
    # When mandatory = True, index should not be None, but we are letting the front-end take care of setting index to 0 because of a bug
    # (see https://github.com/widgetti/solara/issues/282)
    # TODO: set index to 0 on python side (after #282 is resolved)
    index, set_index = solara.use_state_or_update(values.index(reactive_value.value) if reactive_value.value in values else None, key="index")

    def on_index(index):
        set_index(index)
        if mandatory:
            value = values[index]
        else:
            value = values[index] if index is not None else None
        reactive_value.set(value)

    return cast(
        reacton.core.ValueElement[v.BtnToggle, T],
        rv.BtnToggle(children=children, multiple=False, mandatory=mandatory, v_model=index, on_v_model=on_index, dense=dense, class_=class_, style_=style_flat),
    )


@solara.value_component(None)
def ToggleButtonsMultiple(
    value: Union[List[T], solara.Reactive[List[T]]] = [],
    values: List[T] = [],
    children: List[reacton.core.Element] = [],
    on_value: Union[Callable[[List[T]], None], None] = None,
    dense: bool = False,
    mandatory: bool = False,
    classes: List[str] = [],
    style: Union[str, Dict[str, str], None] = None,
) -> reacton.core.ValueElement[v.BtnToggle, List[T]]:
    """ToggleButtons for selecting multiple values.

    ### Basic example:
    ```solara
    import solara

    all_languages = "Python C++ Java JavaScript TypeScript BASIC".split()
    languages = solara.reactive([all_languages[0]])


    @solara.component
    def Page():
        with solara.Card("My favorite programming languages"):
            solara.ToggleButtonsMultiple(languages, all_languages)
            solara.Markdown(f"**Selected**: {languages.value}")
    ```

    ## Arguments

     * `value`: The currently selected values.
     * `values`: List of values to select from.
     * `children`: List of buttons to use as values.
     * `on_value`: Callback to call when the value changes.
     * `dense`: Whether to use a dense (smaller) style.
     * `mandatory`: Whether selecting at least one element is mandatory.
     * `style`: CSS style to apply to the top level element.
     * `classes`: List of CSS classes to be applied to the top level element.
    """
    class_ = _combine_classes(classes)
    style_flat = solara.util._flatten_style(style)
    # See comment regarding typing issue in ToggleButtonsSingle
    reactive_value = solara.use_reactive(value, on_value)  # type: ignore
    children = [solara.Button(label=str(value)) for value in values] + children
    allvalues = values + [_get_button_value(button) for button in children]
    indices, set_indices = solara.use_state_or_update([allvalues.index(k) for k in reactive_value.value], key="index")

    def on_indices(indices):
        set_indices(indices)
        value = [allvalues[k] for k in indices]
        reactive_value.set(value)

    return cast(
        reacton.core.ValueElement[v.BtnToggle, List[T]],
        rv.BtnToggle(
            children=children, multiple=True, mandatory=mandatory, v_model=indices, on_v_model=on_indices, dense=dense, class_=class_, style_=style_flat
        ),
    )



from pathlib import Path
from typing import cast

import playwright
import playwright.sync_api
from reacton.core import _RenderContext

import solara
import solara.server
import solara.server.settings
from solara.server import kernel_context

HERE = Path(__file__).parent


@solara.component
def TwoTexts():
    with solara.Div(classes=["solara-test-div"]):
        solara.Text("AAA")
        solara.Text("BBB")


def test_popout(page_session: playwright.sync_api.Page, solara_server, solara_app, extra_include_path):
    with extra_include_path(HERE), solara_app("popout_test:TwoTexts"):
        page_session.goto(solara_server.base_url + "?solara-no-close-beacon")
        el = page_session.locator("text=AAA")
        assert el.text_content() == "AAA"
        contexts = list(kernel_context.contexts.values())
        assert len(contexts) == 1
        context = contexts[0]
        kernel_id = context.id
        rc = cast(_RenderContext, context.app_object)
        widget = rc.find(children=["BBB"]).widget
        model_id = widget._model_id

        # we should not lose the context, it should be kept alive
        page_session.goto("about:blank")
        page_session.wait_for_timeout(100)
        contexts = list(kernel_context.contexts.values())
        assert len(contexts) == 1

        page_session.goto(solara_server.base_url + f"?kernelid={kernel_id}&modelid={model_id}")
        page_session.locator("text=BBB").wait_for()
        page_session.locator("text=AAA").wait_for(state="detached")

        cull_timeout_previous = solara.server.settings.kernel.cull_timeout
        try:
            solara.server.settings.kernel.cull_timeout = "0s"
            page_session.goto("about:blank")
            page_session.wait_for_timeout(1000)
            contexts = list(kernel_context.contexts.values())
            assert len(contexts) == 0
        finally:
            solara.server.settings.kernel.cull_timeout = cull_timeout_previous



import abc
import logging
from typing import Callable, List, Optional, Tuple, Union, cast

import solara
from solara import _using_solara_server

logger = logging.getLogger("solara.router")


class _LocationBase(abc.ABC):
    @property
    def pathname(self):
        pass

    @pathname.setter
    # mypy does not accept this
    # @abc.abstractmethod
    def pathname(self):
        pass


class _Location(_LocationBase):
    def __init__(self, pathname, setter: Callable[[str], None]) -> None:
        self._pathname = pathname
        self.setter = setter

    @property
    def pathname(self):
        return self._pathname

    @pathname.setter
    def pathname(self, value):
        # import pdb

        # pdb.set_trace()
        self._pathname = value
        self.setter(self._pathname)


class Router:
    search: Optional[str]

    def __init__(self, path: str, routes: List[solara.Route], set_path: Callable[[str], None] = None):
        # see https://developer.mozilla.org/en-US/docs/Web/API/Location for anatomy/nomenclature
        if "?" in path:
            self.path, self.search = path.split("?", 1)
        else:
            self.path = path
            self.search = None
        del path
        self.set_path = set_path
        self.parts = (self.path or "").strip("/").split("/")
        self.routes = routes
        self.root_path = ""
        if _using_solara_server():
            import solara.server.settings

            self.root_path = solara.server.settings.main.root_path or ""
        # each route in this list corresponds to a part in self.parts
        self.path_routes: List["solara.Route"] = []
        self.path_routes_siblings: List[List["solara.Route"]] = []  # siblings including itself
        # routes = routes.copy()
        route = None
        for part in self.parts:
            for route in routes:
                if (route.path == part) or (route.path == "/" and not part):
                    self.path_routes.append(route)
                    self.path_routes_siblings.append(routes)
                    routes = route.children
                    break
        if len(self.parts) == len(self.path_routes):
            # e.g. '/foo/bar' -> ['foo', 'bar'] and bar has a default route
            # but if '' -> [''] we should not
            route = self.path_routes[-1]
            if route:
                default_routes = [k for k in route.children if k.path == "/"]
                if self.parts and self.parts[0] and default_routes:
                    self.path_routes.append(default_routes[0])
                    self.path_routes_siblings.append(route.children)

        assert len(self.path_routes) == len(self.path_routes_siblings)
        self.possible_match = (len(self.path_routes[-1].children) == 0) if self.path_routes else False

    def push(self, path: str):
        assert self.set_path is not None
        self.set_path(path)


router_context = solara.create_context(Router("", []))
_location_context = solara.create_context(cast(_LocationBase, _Location("", lambda x: None)))

route_level_context = solara.create_context(0)


def use_route_level():
    route_level = solara.use_context(route_level_context)
    return route_level


def use_router() -> Router:
    """Returns the current router object.

    See also [Understanding Routing](/docs/understanding/routing).

    `use_router` returns the current router object. This is useful to build custom routing.

    the router object contains the following properties/methods:

     * `path` - the current pathname (e.g. `/fruit/banana`)
     * `parts` - the current pathname split into parts (e.g. `['fruit', 'banana']`)
     * `search` - the query parameters string (e.g. `color=yellow`).
     * `push(path: str)` - navigate to path (e.g. `router.push('/fruit/banana')`)

    ## Typical usage:

    ```python
    import solara


    @solara.component
    def Page():
        router = solara.use_router()

        def redirect():
            router.push(f"/api/use_route")

        solara.Button("Navigate using an event", on_click=redirect)
    ```

    """
    return solara.use_context(router_context)


def use_route(
    level=0,
    peek=False,
) -> Tuple[Optional[solara.Route], List[solara.Route]]:
    """Returns (if found) the current route that matches the pathname, or None

    See also [Understanding Routing](/docs/understanding/routing).

    `use_route` returns (if found) the current route that matches the pathname, or None. It also returns all resolved routes of that level
    (i.e. all siblings and itself). This return tuple is useful to build custom navigation (e.g. using tabs or buttons).


    Routing starts with declaring a set of `routes` in your app (solara picks up the `routes` variable if it exists,
    and it should be in the same namespace as `Page`).
    In the demo below, we declared the following routes.

    ```python
    routes = [
        solara.Route(path="/"),
        solara.Route(
            path="fruit",
            component=Fruit,
            children=[
                solara.Route(path="/"),
                solara.Route(path="kiwi"),
                solara.Route(path="banana"),
                solara.Route(path="apple"),
            ],
        ),
    ]
    ```

    Note that all routes are relative, since a component does not know if it is embedded into a larger application, which may also do routing.
    Therefore you should never use the `route.path` for navigation since the route object has no knowledge of the full url
    (e.g. `/api/use_route/fruit/banana`) but only knows its small piece of the pathname (e.g. `banana`)

    Use [`resolve_path`](/api/resolve_path) to request the full url for navigation, or simply use the `Link` component that can do this for us.

    If the current route has children, any child component that calls `use_route` will return the matched route and its siblings of our children.


    ## Arguments

    * `level`: the level of the route to return. 0 is the current route, -1 is the parent route, 1 the child route, etc.
    * `peek`: if True, the route level is not incremented. This is useful to peek at the next route level without changing the current route level.
    """

    router = solara.use_context(router_context)
    route_level = solara.use_context(route_level_context)
    if not peek:
        route_level_context.provide(route_level + 1)
    route_level += level
    if route_level < len(router.path_routes):
        return router.path_routes[route_level], router.path_routes_siblings[route_level]
    else:
        return None, []


def find_route(path: str) -> Optional[solara.Route]:
    router = solara.get_context(router_context)
    route_level = min(solara.get_context(route_level_context), len(router.path_routes_siblings) - 1)
    for route in router.path_routes_siblings[route_level]:
        if path.startswith(route.path) or (not path and route.path == "/"):
            return route
    return None


def use_pathname():
    location_proxy = solara.use_context(_location_context)

    def setter(value):
        location_proxy.pathname = value

    return location_proxy.pathname, setter


def resolve_path(path_or_route: Union[str, solara.Route], level=0) -> str:
    """Resolve a relative path or a route to an absolute path.

    If the path is a string and starts with a `/'`, it is returned as is.


    ## Typical usage:

    ```python
    ...
    route_current, routes_current_level = solara.routes()
    # route_current.path == "banana"
    path = solara.resolve_path(route_current)
    # path == "/fruit/banana"
    path_same = solara.resolve_path("banana")
    # path_same == path == "/fruit/banana"
    ...
    ```

    ## Arguments

     * path_or_route: a path string or a [`solara.Route`](/api/route) object to resolve.

    ## See also

     * [Multipage](/docs/howto/multipage).
     * [Understanding Routing](/docs/understanding/routing).


    """
    router = solara.get_context(router_context)
    if isinstance(path_or_route, str):
        path = path_or_route
        if path.startswith("/"):
            return path
        route_level = solara.get_context(route_level_context) + level
        parts = [*router.parts[:route_level], path]
        path = "/" + "/".join(parts)
        if path.startswith("//"):
            path = path[1:]
        return path
    elif isinstance(path_or_route, solara.Route):
        route: solara.Route = path_or_route
        path = _resolve_path("/", route, router.routes)
        if path.startswith("//"):
            path = path[1:]
        return path


def _resolve_path(prefix: str, findroute: solara.Route, routes: List[solara.Route]):
    for route in routes:
        path = (prefix + "/" + route.path) if route.path != "/" else prefix
        if findroute is route:
            return path
        possible_path = _resolve_path(path, findroute=findroute, routes=route.children)
        if possible_path is not None:
            return possible_path



import asyncio
from pathlib import Path

import playwright.sync_api

import solara
import solara.server.starlette

HERE = Path(__file__).parent


@solara.component
def Page():
    def run_async():
        async def some_task():
            await asyncio.sleep(0.01)
            label.value = "asyncio run"

        asyncio.create_task(some_task())

    label = solara.use_reactive("initial")
    solara.Button(label.value, on_click=run_async)


def test_async_callback(page_session: playwright.sync_api.Page, solara_app, extra_include_path, solara_server):
    with extra_include_path(HERE), solara_app("async_test"):
        page_session.goto(solara_server.base_url + "/")
        page_session.locator("text=initial").click()
        page_session.locator("text=asyncio run").wait_for()



import threading
from pathlib import Path
from typing import cast

import playwright.sync_api
import pytest
from reacton.core import _RenderContext

import solara.server.kernel_context
import solara.server.server
import solara.server.settings
from solara.server import kernel_context

HERE = Path(__file__).parent


@solara.component
def ClickButton(label="Clicks"):
    clicks = solara.use_reactive(0)
    solara.Button(label=f"{label}-{clicks.value}", on_click=lambda: clicks.set(clicks.value + 1))


@pytest.fixture
def short_cull_timeout():
    cull_timeout_previous = solara.server.settings.kernel.cull_timeout
    solara.server.settings.kernel.cull_timeout = "4.0s"
    try:
        yield
    finally:
        solara.server.settings.kernel.cull_timeout = cull_timeout_previous


def test_kernel_lifecycle_close_single(
    short_cull_timeout,
    browser: playwright.sync_api.Browser,
    page_session: playwright.sync_api.Page,
    solara_server,
    solara_app,
    extra_include_path,
):
    with extra_include_path(HERE), solara_app("lifecycle_test:ClickButton"):
        kernel_context.contexts.clear()
        page_session.goto(solara_server.base_url)
        page_session.locator("text=Clicks-0").click()
        contexts = list(kernel_context.contexts.values())
        assert len(contexts) == 1
        context = contexts[0]
        assert not context.closed_event.is_set()
        page_session.goto("about:blank")
        page_session.wait_for_timeout(100)
        assert context.closed_event.is_set()


def test_kernel_lifecycle_close_while_disconnected(
    short_cull_timeout,
    browser: playwright.sync_api.Browser,
    page_session: playwright.sync_api.Page,
    solara_server,
    solara_app,
    extra_include_path,
):
    with extra_include_path(HERE), solara_app("lifecycle_test:ClickButton"):
        kernel_context.contexts.clear()
        page_session.goto(solara_server.base_url + "?solara-no-close-beacon")
        page_session.locator("text=Clicks-0").click()
        contexts = list(kernel_context.contexts.values())
        assert len(contexts) == 1
        context = contexts[0]
        assert not context.closed_event.is_set()
        page_session.wait_for_timeout(100)
        page_session.goto("about:blank")

        kernel_id = context.id
        rc = cast(_RenderContext, context.app_object)
        widget = rc.container
        assert widget is not None
        model_id = widget._model_id

        page_session.goto(solara_server.base_url + f"?kernelid={kernel_id}&modelid={model_id}")
        # make sure the page is functional
        page_session.locator("text=Clicks-1").click()
        page_session.locator("text=Clicks-2").wait_for()
        page_session.goto("about:blank")
        # give a bit of time to make sure the cull task is started
        page_session.wait_for_timeout(100)
        cull_task_2 = context._last_kernel_cull_task
        assert cull_task_2 is not None
        # we can't mix do async, so we hook up an event to the Future
        event = threading.Event()
        cull_task_2.add_done_callback(lambda x: event.set())
        event.wait()
        assert context.closed_event.is_set()



"""# use_state

"""
import solara
import solara.autorouting
from solara.website.utils import apidoc

from . import NoPage

title = "use_state"
Page = NoPage
__doc__ += apidoc(solara.use_state)  # type: ignore



import asyncio
import dataclasses
import enum
import logging
import os
import pickle
import threading
import time
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, cast

import ipywidgets as widgets
import reacton
from ipywidgets import DOMWidget, Widget

import solara.server.settings
import solara.util

from . import kernel, kernel_context, websocket
from .kernel import Kernel, WebsocketStreamWrapper

WebSocket = Any
logger = logging.getLogger("solara.server.app")


class Local(threading.local):
    kernel_context_stack: Optional[List[Optional["kernel_context.VirtualKernelContext"]]] = None


local = Local()


class PageStatus(enum.Enum):
    CONNECTED = "connected"
    DISCONNECTED = "disconnected"
    CLOSED = "closed"


_on_kernel_start_callbacks: List[Callable[[], Optional[Callable[[], None]]]] = []


def on_kernel_start(f: Callable[[], Optional[Callable[[], None]]]):
    _on_kernel_start_callbacks.append(f)


@dataclasses.dataclass
class VirtualKernelContext:
    id: str
    kernel: kernel.Kernel
    # we keep track of the session id to prevent kernel hijacking
    # to 'steal' a kernel, one would need to know the session id
    # *and* the kernel id
    session_id: str
    control_sockets: List[WebSocket] = dataclasses.field(default_factory=list)
    # this is the 'private' version of the normally global ipywidgets.Widgets.widget dict
    # see patch.py
    widgets: Dict[str, Widget] = dataclasses.field(default_factory=dict)
    # same, for ipyvue templates
    # see patch.py
    templates: Dict[str, Widget] = dataclasses.field(default_factory=dict)
    user_dicts: Dict[str, Dict] = dataclasses.field(default_factory=dict)
    # anything we need to attach to the context
    # e.g. for a react app the render context, so that we can store/restore the state
    app_object: Optional[Any] = None
    reload: Callable = lambda: None
    state: Any = None
    container: Optional[DOMWidget] = None
    # we track which pages are connected to implement kernel culling
    page_status: Dict[str, PageStatus] = dataclasses.field(default_factory=dict)
    # only used for testing
    _last_kernel_cull_task: "Optional[asyncio.Future[None]]" = None
    closed_event: threading.Event = dataclasses.field(default_factory=threading.Event)
    _on_close_callbacks: List[Callable[[], None]] = dataclasses.field(default_factory=list)

    def __post_init__(self):
        with self:
            for f in _on_kernel_start_callbacks:
                cleanup = f()
                if cleanup:
                    self.on_close(cleanup)

    def restart(self):
        # should we do this, or maybe close the context and create a new one?
        with self:
            for f in reversed(self._on_close_callbacks):
                f()
            self._on_close_callbacks.clear()
            self.__post_init__()

    def display(self, *args):
        print(args)  # noqa

    def on_close(self, f: Callable[[], None]):
        self._on_close_callbacks.append(f)

    def __enter__(self):
        if local.kernel_context_stack is None:
            local.kernel_context_stack = []
        key = get_current_thread_key()
        local.kernel_context_stack.append(current_context.get(key, None))
        current_context[key] = self

    def __exit__(self, *args):
        key = get_current_thread_key()
        assert local.kernel_context_stack is not None
        current_context[key] = local.kernel_context_stack.pop()

    def close(self):
        logger.info("Shut down virtual kernel: %s", self.id)
        with self:
            for f in reversed(self._on_close_callbacks):
                f()
        with self:
            if self.app_object is not None:
                if isinstance(self.app_object, reacton.core._RenderContext):
                    try:
                        self.app_object.close()
                    except Exception as e:
                        logger.exception("Could not close render context: %s", e)
                        # we want to continue, so we at least close all widgets
            widgets.Widget.close_all()
            # what if we reference each other
            # import gc
            # gc.collect()
            self.kernel.session.close()
        if self.id in contexts:
            del contexts[self.id]
        self.closed_event.set()

    def _state_reset(self):
        state_directory = Path(".") / "states"
        state_directory.mkdir(exist_ok=True)
        path = state_directory / f"{self.id}.pickle"
        path = path.absolute()
        try:
            path.unlink()
        except:  # noqa
            pass
        del contexts[self.id]
        key = get_current_thread_key()
        del current_context[key]

    def state_save(self, state_directory: os.PathLike):
        path = Path(state_directory) / f"{self.id}.pickle"
        render_context = self.app_object
        if render_context is not None:
            render_context = cast(reacton.core._RenderContext, render_context)
            state = render_context.state_get()
            with path.open("wb") as f:
                logger.debug("State: %r", state)
                pickle.dump(state, f)

    def page_connect(self, page_id: str):
        logger.info("Connect page %s for kernel %s", page_id, self.id)
        assert self.page_status.get(page_id) != PageStatus.CLOSED, "cannot connect with the same page_id after a close"
        self.page_status[page_id] = PageStatus.CONNECTED
        if self._last_kernel_cull_task:
            self._last_kernel_cull_task.cancel()

    def page_disconnect(self, page_id: str) -> "asyncio.Future[None]":
        """Signal that a page has disconnected, and schedule a kernel cull if needed.

        During the kernel reconnect window, we will keep the kernel alive, even if all pages have disconnected.

        Returns a future that is set when the kernel cull is done.
        The scheduled kernel cull can be cancelled when a new page connects, a new disconnect is scheduled,
        or a page if explicitly closed.
        """
        logger.info("Disconnect page %s for kernel %s", page_id, self.id)
        future: "asyncio.Future[None]" = asyncio.Future()
        self.page_status[page_id] = PageStatus.DISCONNECTED
        current_event_loop = asyncio.get_event_loop()

        async def kernel_cull():
            try:
                cull_timeout_sleep_seconds = solara.util.parse_timedelta(solara.server.settings.kernel.cull_timeout)
                logger.info("Scheduling kernel cull, will wait for max %s before shutting down the virtual kernel %s", cull_timeout_sleep_seconds, self.id)
                await asyncio.sleep(cull_timeout_sleep_seconds)
                has_connected_pages = PageStatus.CONNECTED in self.page_status.values()
                if has_connected_pages:
                    logger.info("We have (re)connected pages, keeping the virtual kernel %s alive", self.id)
                else:
                    logger.info("No connected pages, and timeout reached, shutting down virtual kernel %s", self.id)
                    self.close()
                current_event_loop.call_soon_threadsafe(future.set_result, None)
            except asyncio.CancelledError:
                current_event_loop.call_soon_threadsafe(future.cancel, "cancelled because a new cull task was scheduled")
                raise

        has_connected_pages = PageStatus.CONNECTED in self.page_status.values()
        if not has_connected_pages:
            # when we have no connected pages, we will schedule a kernel cull
            if self._last_kernel_cull_task:
                self._last_kernel_cull_task.cancel()

            async def create_task():
                task = asyncio.create_task(kernel_cull())
                # create a reference to the task so we can cancel it later
                self._last_kernel_cull_task = task
                await task

            asyncio.run_coroutine_threadsafe(create_task(), keep_alive_event_loop)
        else:
            future.set_result(None)
        return future

    def page_close(self, page_id: str):
        """Signal that a page has closed, and close the context if needed.

        Closing the browser tab or a page navigation means an explicit close, which is
        different from a websocket/page disconnect, which we might want to recover from.

        """
        self.page_status[page_id] = PageStatus.CLOSED
        logger.info("Close page %s for kernel %s", page_id, self.id)
        has_connected_pages = PageStatus.CONNECTED in self.page_status.values()
        has_disconnected_pages = PageStatus.DISCONNECTED in self.page_status.values()
        if not (has_connected_pages or has_disconnected_pages):
            logger.info("No connected or disconnected pages, shutting down virtual kernel %s", self.id)
            if self._last_kernel_cull_task:
                self._last_kernel_cull_task.cancel()
            self.close()


try:
    # Normal Python
    keep_alive_event_loop = asyncio.new_event_loop()

    def _run():
        asyncio.set_event_loop(keep_alive_event_loop)
        try:
            keep_alive_event_loop.run_forever()
        except Exception:
            logger.exception("Error in keep alive event loop")
            raise

    threading.Thread(target=_run, daemon=True).start()
except RuntimeError:
    # Emscripten/pyodide/lite
    keep_alive_event_loop = asyncio.get_event_loop()

contexts: Dict[str, VirtualKernelContext] = {}
# maps from thread key to VirtualKernelContext, if VirtualKernelContext is None, it exists, but is not set as current
current_context: Dict[str, Optional[VirtualKernelContext]] = {}


def create_dummy_context():
    from . import kernel

    kernel_context = VirtualKernelContext(
        id="dummy",
        session_id="dummy",
        kernel=kernel.Kernel(),
    )
    return kernel_context


def get_current_thread_key() -> str:
    thread = threading.current_thread()
    return get_thread_key(thread)


def get_thread_key(thread: threading.Thread) -> str:
    thread_key = thread._name + str(thread._ident)  # type: ignore
    return thread_key


def set_context_for_thread(context: VirtualKernelContext, thread: threading.Thread):
    key = get_thread_key(thread)
    current_context[key] = context


def has_current_context() -> bool:
    thread_key = get_current_thread_key()
    return (thread_key in current_context) and (current_context[thread_key] is not None)


def get_current_context() -> VirtualKernelContext:
    thread_key = get_current_thread_key()
    if thread_key not in current_context:
        raise RuntimeError(
            f"Tried to get the current context for thread {thread_key}, but no known context found. This might be a bug in Solara. "
            f"(known contexts: {list(current_context.keys())}"
        )
    context = current_context[thread_key]
    if context is None:
        raise RuntimeError(
            f"Tried to get the current context for thread {thread_key!r}, although the context is know, it was not set for this thread. "
            + "This might be a bug in Solara."
        )
    return context


def set_current_context(context: Optional[VirtualKernelContext]):
    thread_key = get_current_thread_key()
    current_context[thread_key] = context


def initialize_virtual_kernel(session_id: str, kernel_id: str, websocket: websocket.WebsocketWrapper):
    from solara.server import app as appmodule

    if kernel_id in contexts:
        logger.info("reusing virtual kernel: %s", kernel_id)
        context = contexts[kernel_id]
        if context.session_id != session_id:
            logger.critical("Session id mismatch when reusing kernel (hack attempt?): %s != %s", context.session_id, session_id)
            websocket.send_text("Session id mismatch when reusing kernel (hack attempt?)")
            # to avoid very fast reconnects (we are in a thread anyway)
            time.sleep(0.5)
            raise ValueError("Session id mismatch")
        kernel = context.kernel
    else:
        kernel = Kernel()
        logger.info("new virtual kernel: %s", kernel_id)
        context = contexts[kernel_id] = VirtualKernelContext(id=kernel_id, session_id=session_id, kernel=kernel, control_sockets=[], widgets={}, templates={})

        with context:
            widgets.register_comm_target(kernel)
            appmodule.register_solara_comm_target(kernel)
    with context:
        assert kernel is Kernel.instance()
        kernel.shell_stream = WebsocketStreamWrapper(websocket, "shell")
        kernel.control_stream = WebsocketStreamWrapper(websocket, "control")
        kernel.session.websockets.add(websocket)
    return context



"""
# InputDate

This page contains the two variants of datepickers available in solara

# InputDate
"""
import solara
from solara.website.utils import apidoc

from . import NoPage

title = "InputDate"


__doc__ += apidoc(solara.lab.components.input_date.InputDate.f)  # type: ignore
__doc__ += "# InputDateRange"
__doc__ += apidoc(solara.lab.components.input_date.InputDateRange.f)  # type: ignore

Page = NoPage



"""
# AvatarMenu

"""
try:
    from solara_enterprise import auth
except ImportError:
    auth = None
from solara.website.utils import apidoc

from . import NoPage

title = "AvatarMenu"

Page = NoPage


if auth:
    __doc__ += apidoc(auth.AvatarMenu.f)  # type: ignore
else:
    __doc__ += "solara-enterprise not installed."



from typing import Callable, Dict, Union, cast

import ipyvuetify.Themes
from ipyvuetify.Themes import Theme

import solara
from solara.components.component_vue import component_vue
from solara.tasks import Proxy

theme = Proxy(Theme)
ipyvuetify.Themes.theme = cast(ipyvuetify.Themes.Theme, theme)


def use_dark_effective():
    """Return True if the frontend is using a dark theme.

    Equivalent of

    ```python
    solara.use_trait_observe(solara.lab.theme, "dark_effective")
    ```

    See [use_trait_observe](/api/use_trait_observe).
    """
    return solara.use_trait_observe(theme, "dark_effective")


def _set_theme(themes: Union[Dict[str, Dict[str, str]], None]):
    if themes is None:
        return

    for theme_type in themes.keys():
        widget = getattr(theme.themes, theme_type)
        with widget.hold_trait_notifications():
            for k, v in themes[theme_type].items():
                setattr(widget, k, v)


def _get_theme(theme: Theme) -> Dict[str, Dict[str, str]]:
    theme_dict: Dict[str, Dict[str, str]] = cast(Dict[str, Dict[str, str]], {})
    for theme_type, theme_value in theme.themes.__dict__.items():
        theme_traits = theme_value.keys
        theme_dict[theme_type] = {}
        for trait in theme_traits:
            if not trait.startswith("_"):
                theme_dict[theme_type][trait] = getattr(theme_value, trait)
    return theme_dict


@component_vue("theming.vue")
def _ThemeToggle(
    theme_dark: str,
    event_sync_themes: Callable[[str], None],
    enable_auto: bool,
    on_icon: str,
    off_icon: str,
    auto_icon: str,
    clicks: int = 1,
):
    pass


@solara.component
def ThemeToggle(
    on_icon: str = "mdi-weather-night",
    off_icon: str = "mdi-weather-sunny",
    auto_icon: str = "mdi-brightness-auto",
    enable_auto: bool = True,
):
    """
    Insert a toggle switch for user to switch between light and dark themes.

    ```solara
    import solara.lab

    @solara.component
    def Page():
        solara.lab.ThemeToggle()
    ```

    ## Arguments
    - `on_icon`: The icon to display when the dark theme is enabled.
    - `off_icon`: The icon to display when the dark theme is disabled.
    - `auto_icon`: The icon to display when the theme is set to auto. Only visible if `enable_auto` is `True`.
    - `enable_auto`: Whether to enable the auto detection of dark mode.
    """

    def sync_themes(selected_theme: str):
        theme.dark = selected_theme

    return _ThemeToggle(
        theme_dark=theme.dark,
        event_sync_themes=sync_themes,
        enable_auto=enable_auto,
        on_icon=on_icon,
        off_icon=off_icon,
        auto_icon=auto_icon,
    )



"""# Linked views

This example shows how to create linked views. The clicked row information is passed up to the parent component using an event,
and then passed down to both child components.

"""

import dataclasses
from typing import Callable, cast

import plotly.express as px

import solara

df = px.data.iris()
columns = list(df.columns)


@dataclasses.dataclass
class ClickPoint:
    row_index: int
    x_column: str
    y_column: str


def find_row_index(fig, click_data):
    # goes from trace index and point index to row index in a dataframe
    # requires passing df.index as to custom_data
    trace_index = click_data["points"]["trace_indexes"][0]
    point_index = click_data["points"]["point_indexes"][0]
    trace = fig.data[trace_index]
    return trace.customdata[point_index][0]


@solara.component
def ClickScatter(df, x, y, color, click_row, on_click: Callable[[ClickPoint], None]):
    x, set_x = solara.use_state(x)
    y, set_y = solara.use_state(y)
    fig = px.scatter(df, x, y, color=color, custom_data=[df.index])

    def on_click_trace(click_data):
        # sanity checks
        assert click_data["event_type"] == "plotly_click"
        row_index = find_row_index(fig, click_data)
        on_click(ClickPoint(row_index, x, y))

    if click_row:
        click_x = df[x].values[click_row]
        click_y = df[y].values[click_row]
        fig.add_trace(px.scatter(x=[click_x], y=[click_y], text=["⭐️"]).data[0])
    # make the figure a bit smaller
    fig.update_layout(width=400)
    with solara.VBox() as main:
        solara.FigurePlotly(fig, on_click=on_click_trace)
        solara.Select(label="X-axis", value=x, values=columns, on_value=set_x)
        solara.Select(label="Y-axis", value=y, values=columns, on_value=set_y)
    return main


@solara.component
def Page():
    click_point, set_click_point = solara.use_state(cast(ClickPoint, None))
    if click_point:
        clicked_row = click_point.row_index
    else:
        clicked_row = None

    with solara.VBox() as main:
        with solara.HBox():
            ClickScatter(df, "sepal_length", "sepal_width", "species", clicked_row, on_click=set_click_point)
            ClickScatter(df, "petal_length", "petal_width", "species", clicked_row, on_click=set_click_point)
        if click_point is not None:
            clicked_row = click_point.row_index
            solara.Success(f"Clicked on row {clicked_row}. Which is highlighted in the both plots.")
            solara.Markdown(
                f"""
            ```python
            row_data = {df.iloc[clicked_row].to_dict()}
            ```"""
            )
        else:
            solara.Info("Click to select a point")

    return main



"""
# FileDrop
"""
import textwrap

import solara
from solara.components.file_drop import FileInfo
from solara.website.utils import apidoc


@solara.component
def Page():
    content, set_content = solara.use_state(b"")
    filename, set_filename = solara.use_state("")
    size, set_size = solara.use_state(0)

    def on_file(file: FileInfo):
        set_filename(file["name"])
        set_size(file["size"])
        f = file["file_obj"]
        set_content(f.read(100))

    with solara.Div() as main:
        solara.FileDrop(
            label="Drag and drop a file here to read the first 100 bytes",
            on_file=on_file,
            lazy=True,  # We will only read the first 100 bytes
        )
        if content:
            solara.Info(f"File {filename} has total length: {size}\n, first 100 bytes:")
            solara.Preformatted("\n".join(textwrap.wrap(repr(content))))

    return main


__doc__ += apidoc(solara.FileDrop.f)  # type: ignore



import ast
import linecache
from pathlib import Path
from typing import Any, Dict

import IPython
import nbformat

import solara
import solara.components.applayout
import solara.toestand
from solara.components.markdown import ExceptionGuard

HERE = Path(__file__).parent


def execute_notebook(path: Path):
    nb: nbformat.NotebookNode = nbformat.read(path, 4)
    stat = path.stat()
    scope: Dict[str, Any] = {}
    for cell_index, cell in enumerate(nb.cells):
        cell_index += 1  # used 1 based
        cell.scope_snapshot = {}
        if cell.cell_type == "code":
            if cell.source.startswith("## solara: skip"):
                continue

            def execute(cell=cell):
                source = cell.source
                cell_path = f"app.ipynb input cell {cell_index}"
                entry = (
                    stat.st_size,
                    stat.st_mtime,
                    [line + "\n" for line in source.splitlines()],
                    cell_path,
                )

                linecache.cache[cell_path] = entry
                code = ast.parse(source, cell_path, "exec")
                value = None
                with solara.Column():
                    if code.body:
                        last_statement = code.body[-1]
                        if isinstance(last_statement, ast.Expr):
                            code.body.pop()
                            last = ast.Expression(last_statement.value)
                            exec(compile(code, "<string>", mode="exec"), scope)
                            value = eval(compile(last, "<string>", mode="eval"), scope)
                        else:
                            exec(compile(code, "<string>", mode="exec"), scope)
                    import matplotlib_inline.backend_inline

                    matplotlib_inline.backend_inline.flush_figures()
                scope_snapshot = {**scope}
                cell.__dict__["scope_snapshot"] = scope_snapshot
                return value

            # every cell gets a snapshot of the scope
            # avoid the buggy setitem of NotebookNode
            # which causes an recursion error
            yield cell_index, cell, execute
        elif cell.cell_type == "markdown":
            # last_expressions.append(None)
            yield cell_index, cell, lambda: None
        else:
            raise ValueError(f"Unknown cell type: {cell.cell_type}, supported types are: code, markdown")


@solara.component
def NotebookExecute(notebook_path: Path, show_last_expressions=False, auto_show_page=False):
    import IPython.core.pylabtools as pylabtools

    shell = IPython.get_ipython().kernel.shell

    # TODO: there is a change the cleanup will not be called
    pylabtools.select_figure_formats(shell, ["png"])

    def cleanup():
        def _cleanup():
            from matplotlib.figure import Figure

            png_formatter = shell.display_formatter.formatters["image/png"]
            png_formatter.pop(Figure)

        return _cleanup

    solara.use_effect(cleanup, [])
    values = list(solara.use_memo(lambda: execute_notebook(notebook_path)))

    last_page = None

    solara.components.applayout.should_use_embed.provide(True)
    with solara.Column(style={"max-width": "100%"}) as main:

        for cell_index, cell, executor in values:
            cell_index += 1  # used 1 based
            if cell.cell_type == "code":
                if cell.source.startswith("## solara: skip"):
                    continue
                solara.Markdown(
                    f"""
```python
{cell.source}
```"""
                )
                # we execute here, such that any side effect (display, etc)
                # will be done under the context of the Column
                with ExceptionGuard():
                    with solara.AppLayout(navigation=False, toolbar_dark=True):
                        # we don't want to listen to reactive variables
                        copy = set(solara.toestand.thread_local.reactive_used or [])
                        try:
                            last_expression = executor()
                        finally:
                            solara.toestand.thread_local.reactive_used = copy
                    scope = cell.scope_snapshot
                    page = scope.get("Page")
                    if auto_show_page and page != last_page and page is not None:
                        page()
                    else:
                        if show_last_expressions:
                            if last_expression is not None:
                                solara.display(last_expression)
                last_page = page
            elif cell.cell_type == "markdown":
                solara.Markdown(cell.source)
            else:
                raise ValueError(f"Unknown cell type: {cell.cell_type}, supported types are: code, markdown")
    return main


@solara.component
def Notebook(notebook_path: Path, show_last_expressions=False, auto_show_page=False, execute=True, outputs={}):
    if execute:
        return NotebookExecute(notebook_path, show_last_expressions, auto_show_page)
    else:
        with solara.Column(style={"max-width": "100%"}) as main:
            nb: nbformat.NotebookNode = nbformat.read(notebook_path, 4)
            for cell_index, cell in enumerate(nb.cells):
                cell_index += 1
                if cell.cell_type == "code":
                    if cell.source.startswith("## solara: skip"):
                        continue
                    solara.Markdown(
                        f"""
```python
{cell.source}
```"""
                    )
                    if cell.outputs:
                        for output in cell.outputs:
                            if output.output_type == "display_data":
                                solara.display(output.data, raw=True)
                            elif output.output_type == "execute_result":
                                if "text/html" in output.data:
                                    solara.display(output.data, raw=True)
                    else:
                        cell_id = str(cell.id)
                        output = outputs.get(cell_id) or outputs.get(cell_index)
                        if output is None:
                            if cell_id in outputs or cell_index in outputs:
                                pass  # explicit None
                            else:
                                solara.display(f"Output missing for cell: {cell_index} id {cell_id}")
                        else:
                            solara.display(output)

                elif cell.cell_type == "markdown":
                    solara.Markdown(cell.source)
                else:
                    raise ValueError(f"Unknown cell type: {cell.cell_type}, supported types are: code, markdown")
        return main



import base64
import contextlib
import hashlib
import os
import sys
import threading
from collections import abc
from pathlib import Path
from typing import TYPE_CHECKING, Dict, List, Tuple, Union

if TYPE_CHECKING:
    import numpy as np

import ipyvuetify
import ipywidgets
import reacton

import solara

SOLARA_ALLOW_OTHER_TRACER = os.environ.get("SOLARA_ALLOW_OTHER_TRACER", False) in (True, "True", "true", "1")
ipyvuetify_major_version = int(ipyvuetify.__version__.split(".")[0])
ipywidgets_major = int(ipywidgets.__version__.split(".")[0])


def github_url(file):
    rel_path = os.path.relpath(file, Path(solara.__file__).parent.parent)
    github_url = solara.github_url + f"/blob/{solara.git_branch}/" + rel_path
    return github_url


def github_edit_url(file):
    # e.g. https://github.com/widgetti/solara/edit/master/solara/__init__.py
    rel_path = os.path.relpath(file, Path(solara.__file__).parent.parent)
    github_url = solara.github_url + f"/edit/{solara.git_branch}/" + rel_path
    return github_url


def load_file_as_data_url(file_name, mime):
    with open(file_name, "rb") as f:
        data = f.read()
    return f"data:{mime};base64," + base64.b64encode(data).decode("utf-8")


def isinstanceof(object, spec: str):
    """Check if object is instance of type '<modulename>:<classname>'

    This can avoid a runtime dependency, since we do not need to import `modulename`.

    >>> import numpy as np
    >>> isinstanceof(np.arange(2), "numpy:ndarray")
    True
    """
    module_name, classname = spec.split(":")
    module = sys.modules.get(module_name)
    if module:
        cls = getattr(module, classname)
        return isinstance(object, cls)
    return False


def numpy_to_image(data: "np.ndarray", format="png"):
    import io

    if data.ndim == 3:
        try:
            import PIL.Image
        except ModuleNotFoundError:
            raise ModuleNotFoundError("Pillow is required to convert numpy array to image, use pip install pillow to install it.")
        if data.shape[2] == 3:
            im = PIL.Image.fromarray(data[::], "RGB")
        elif data.shape[2] == 4:
            im = PIL.Image.fromarray(data[::], "RGBA")
        else:
            raise ValueError(f"Expected last dimension to have 3 or 4 dimensions, total shape we got was {data.shape}")
        f = io.BytesIO()
        im.save(f, format)
        return f.getvalue()
    else:
        raise ValueError(f"Expected an image with 3 dimensions (height, width, channel), not {data.shape}")


@contextlib.contextmanager
def cwd(path):
    cwd = os.getcwd()
    try:
        os.chdir(path)
        yield
    finally:
        os.chdir(cwd)


def numpy_equals(a, b):
    import numpy as np

    if a is b:
        return True
    if a is None or b is None:
        return False
    if np.all(a == b):
        return True
    return False


def _combine_classes(class_list: List[str]) -> str:
    return " ".join(class_list)


def _flatten_style(style: Union[str, Dict, None] = None) -> str:
    if style is None:
        return ""
    elif isinstance(style, str):
        return style
    elif isinstance(style, dict):
        return ";".join(f"{k}:{v}" for k, v in style.items()) + ";"
    else:
        raise ValueError(f"Expected style to be a string or dict, got {type(style)}")


def import_item(name: str):
    """Import an object by name like solara.cache.LRU"""
    parts = name.rsplit(".", 2)
    if len(parts) == 1:
        return __import__(name)
    else:
        module = __import__(".".join(parts[:-1]), fromlist=[parts[-1]])
        return getattr(module, parts[-1])


def get_solara_home() -> Path:
    """Get solara home directory, defaults to ~/.solara.

    The $SOLARA_HOME environment variable can be set to override this default.

    If $SOLARA_HOME is not defined and ~ cannot be expanded, the current working directory + ".solara" is used.
    """
    os_home = None
    try:
        os_home = Path.home()
    except Exception:
        pass
    if "SOLARA_HOME" in os.environ:
        return Path(os.environ["SOLARA_HOME"])
    elif os_home:
        return os_home / ".solara"
    else:
        return Path(os.getcwd()) / ".solara"


def parse_size(size: str) -> int:
    """Given a human readable size, return the number of bytes.

    Supports GB, MB, KB, and bytes. E.g. 10GB, 500MB, 1KB, 1000

    Commas and _ are ignored, e.g. 1,000,000 is the same as 1000000.
    """
    size = size.replace(",", "").replace("_", "").upper()
    if size.endswith("GB"):
        return int(float(size[:-2]) * 1024 * 1024 * 1024)
    elif size.endswith("MB"):
        return int(float(size[:-2]) * 1024 * 1024)
    elif size.endswith("KB"):
        return int(float(size[:-2]) * 1024)
    elif size.endswith("B"):
        return int(float(size[:-1]))
    else:
        return int(size)


def nested_get(object, dotted_name: str, default=None):
    names = dotted_name.split(".")
    for name in names:
        if isinstance(object, abc.Mapping):
            if name == names[-1]:
                object = object.get(name, default)
            else:
                object = object.get(name)
        else:
            if name == names[-1]:
                object = getattr(object, name, default)
            else:
                object = getattr(object, name)
    return object


# inherit from BaseException so less change of being caught
# in an except
class CancelledError(BaseException):
    pass


# not available in python 3.6
class nullcontext(contextlib.AbstractContextManager):
    def __init__(self, enter_result=None):
        self.enter_result = enter_result

    def __enter__(self):
        return self.enter_result

    def __exit__(self, *excinfo):
        pass


@contextlib.contextmanager
def cancel_guard(cancelled: threading.Event):
    def tracefunc(frame, event, arg):
        # this gets called at least for every line executed
        if cancelled.is_set():
            rc = reacton.core.get_render_context(required=False)
            # we do not want to cancel the rendering cycle
            if rc is None or not rc._is_rendering:
                # this will bubble up
                raise CancelledError()
        if prev and SOLARA_ALLOW_OTHER_TRACER:
            prev(frame, event, arg)
        # keep tracing:
        return tracefunc

    # see https://docs.python.org/3/library/sys.html#sys.settrace
    # it is for the calling thread only
    # not every Python implementation has it
    prev = None
    if hasattr(sys, "gettrace"):
        prev = sys.gettrace()
    if hasattr(sys, "settrace"):
        sys.settrace(tracefunc)
    try:
        yield
    finally:
        if hasattr(sys, "settrace"):
            sys.settrace(prev)


def parse_timedelta(size: str) -> float:
    """Turn a human readable time delta into seconds.
    Supports days(d), hours (h), minutes (m) and seconds (s).
    If not unit is specified, seconds is assumed.
    >>> parse_timedelta("1d")
    86400
    >>> parse_timedelta("1h")
    3600
    >>> parse_timedelta("30m")
    1800
    >>> parse_timedelta("10s")
    10
    >>> parse_timedelta("10")
    10
    """
    if size.endswith("d"):
        return float(size[:-1]) * 24 * 60 * 60
    elif size.endswith("h"):
        return float(size[:-1]) * 60 * 60
    elif size.endswith("m"):
        return float(size[:-1]) * 60
    elif size.endswith("s"):
        return float(size[:-1])
    else:
        return float(size)


def get_file_hash(path: Path, algorithm="md5") -> Tuple[bytes, str]:
    """Compute the hash of a file. Note that we also return the file content as bytes."""
    data = path.read_bytes()
    if sys.version_info[:2] < (3, 9):
        # usedforsecurity is only available in Python 3.9+
        h = hashlib.new(algorithm)
    else:
        h = hashlib.new(algorithm, usedforsecurity=False)  # type: ignore
    h.update(data)
    return data, h.hexdigest()


def is_running_in_colab():
    try:
        import google.colab  # noqa

        return True
    except ImportError:
        return False


def is_running_in_vscode():
    return "VSCODE_PID" in os.environ or "VSCODE_CWD" in os.environ


def is_running_in_voila():
    return os.environ.get("SERVER_SOFTWARE", "").startswith("voila")



from unittest.mock import Mock

import IPython.display
import ipywidgets as widgets

from solara.server import kernel, kernel_context


def test_interactive_shell(no_kernel_context):
    ws1 = Mock()
    ws2 = Mock()
    kernel1 = kernel.Kernel()
    kernel2 = kernel.Kernel()
    kernel1.session.websockets.add(ws1)
    kernel2.session.websockets.add(ws2)
    context1 = kernel_context.VirtualKernelContext(id="1", kernel=kernel1, session_id="session-1")
    context2 = kernel_context.VirtualKernelContext(id="2", kernel=kernel2, session_id="session-2")

    with context1:
        output1 = widgets.Output()
        with output1:
            IPython.display.display("test1")
        assert output1.outputs[0]["data"]["text/plain"] == "'test1'"
        assert ws1.send.call_count == 3  # create 2 widgets (layout and output) and update data
        assert ws2.send.call_count == 0
    with context2:
        output2 = widgets.Output()
        with output2:
            IPython.display.display("test2")
        assert output2.outputs[0]["data"]["text/plain"] == "'test2'"
        assert ws1.send.call_count == 3
        assert ws2.send.call_count == 3

    context1.close()
    context2.close()


def test_clear_output():
    output1 = widgets.Output()
    with output1:
        IPython.display.display("test1")
        IPython.display.clear_output()
    assert len(output1.outputs) == 0



"""
# GridDraggable

Child components are laid out on a grid, which can be dragged and resized.
"""
import pprint
import textwrap

import solara

from .common import ColorCard

title = "GridDraggable"


@solara.component
def Page():
    grid_layout_initial = [
        {"h": 5, "i": "0", "moved": False, "w": 3, "x": 0, "y": 0},
        {"h": 5, "i": "1", "moved": False, "w": 5, "x": 3, "y": 0},
        {"h": 11, "i": "2", "moved": False, "w": 4, "x": 8, "y": 0},
        {"h": 12, "i": "3", "moved": False, "w": 5, "x": 0, "y": 5},
        {"h": 6, "i": "4", "moved": False, "w": 3, "x": 5, "y": 5},
        {"h": 6, "i": "5", "moved": False, "w": 7, "x": 5, "y": 11},
    ]

    colors = "green red orange brown yellow pink".split()

    # we need to store the state of the grid_layout ourselves, otherwise it will 'reset'
    # each time we change resizable or draggable
    grid_layout, set_grid_layout = solara.use_state(grid_layout_initial)

    # some placeholders
    items = [ColorCard(title=f"Child {i}", color=colors[i]) for i in range(len(grid_layout))]
    with solara.VBox() as main:
        resizable = solara.ui_checkbox("Allow resizing", value=True)
        draggable = solara.ui_checkbox("Allow dragging", value=True)

        def reset_layout():
            set_grid_layout(grid_layout_initial)

        solara.Button("Reset to initial layout", on_click=reset_layout)

        solara.GridDraggable(items=items, grid_layout=grid_layout, resizable=resizable, draggable=draggable, on_grid_layout=set_grid_layout)

        # some string kung fu to make this print nicely
        grid_layout_formatted = pprint.pformat(grid_layout, indent=4)
        grid_layout_formatted = textwrap.indent(grid_layout_formatted, " " * len("grid_layout = "))
        grid_layout_formatted = grid_layout_formatted[len("grid_layout = ") :]
        solara.Markdown(
            f"""
            # Resulting layout

            This layout can be copy pasted to put in your code as an initial layout:

            ```python
            grid_layout = {grid_layout_formatted}
            ```
            """
        )
    return main



from pathlib import Path
from typing import BinaryIO, Callable, Optional, Union, cast

import ipyvuetify as vy
import ipywidgets as widgets
import traitlets

import solara


class FileDownloadWidget(vy.VuetifyTemplate):
    template_file = (__file__, "download.vue")
    children = traitlets.List().tag(sync=True, **widgets.widget_serialization)
    filename = traitlets.Unicode().tag(sync=True)
    bytes = traitlets.Bytes(None, allow_none=True).tag(sync=True)
    mime_type = traitlets.Unicode("application/octet-stream").tag(sync=True)
    request_download = traitlets.Bool(False).tag(sync=True)


@solara.component
def FileDownload(
    data: Union[str, bytes, BinaryIO, Callable[[], Union[str, bytes, BinaryIO]]],
    filename: Optional[str] = None,
    label: Optional[str] = None,
    icon_name: Optional[str] = "mdi-cloud-download-outline",
    close_file: bool = True,
    mime_type: str = "application/octet-stream",
    string_encoding: str = "utf-8",
    children=[],
):
    """Download a file or data.

    ## Simple usage

    By default, if no children are provided, a button is created with the label "Download: {filename}".

    ```solara
    import solara

    data = "This is the content of the file"

    @solara.component
    def Page():
        solara.FileDownload(data, filename="solara-download.txt", label="Download file")
    ```

    ## Advanced usage

    If children are provided, they are displayed instead of the button. The children can be any solara component,
    including a button, markdown text, or an image.

    ```solara
    import solara

    data = "This is the content of the file"

    @solara.component
    def Page():
        with solara.FileDownload(data, "solara-download-2.txt"):
            solara.Markdown("Any text, or even an image")
            solara.Image("https://solara.dev/static/public/beach.jpeg", width="200px")
    ```

    ## Custom button

    If children are provided, they are displayed instead of the button. The children can be any solara component,
    including a button, markdown text, or an image.

    ```solara
    import solara

    data = "This is the content of the file"

    @solara.component
    def Page():
        with solara.FileDownload(data, "solara-download-2.txt"):
            solara.Button("Custom download button", icon_name="mdi-cloud-download-outline", color="primary")
    ```

    ## Usage with file

    A file object can be used as data. The file will be closed after downloading by default.

    ```solara
    import solara
    import pandas as pd

    df = pd.DataFrame({"id": [1, 2, 3], "name": ["John", "Mary", "Bob"]})

    @solara.component
    def Page():
        file_object = df.to_csv(index=False)
        solara.FileDownload(file_object, "users.csv", mime_type="application/vnd.ms-excel")
    ```

    If a file like object is used, we try to base the filename on the file object.
    ```solara
    import solara
    import solara.website.pages
    import os

    filename = os.path.dirname(solara.website.__file__) + "/public/beach.jpeg"

    @solara.component
    def Page():
        # only open the file once by using use_memo
        file_object = solara.use_memo(lambda: open(filename, "rb"), [])
        # no filename is provided, but we can extract it from the file object
        solara.FileDownload(file_object, mime_type="image/jpeg", close_file=False)
    ```

    ## Lazy reading

    Not only is the data lazily uploaded to the browser, but also the data is only read when the download is requested.
    This happens for files by default, but can also be used by passing in a callback function.

    ```solara
    import solara
    import time

    @solara.component
    def Page():
        def get_data():
            # I run in a thread, so I can do some heavy processing
            time.sleep(3)
            # I only get called when the download is requested
            return "This is the content of the file"
        solara.FileDownload(get_data, "solara-lazy-download.txt")
    ```

    ## Arguments

     * `data`: The data to download. Can be a string, bytes, or a file like object, or a function that returns one of these.
     * `filename`: The name of the file the user will see as default when downloading (default name is "solara-download.dat").
        If a file object is provided, the filename will be extracted from the file object if possible.
     * `label`: The label of the button. If not provided, the label will be "Download: {filename}".
     - `icon_name`: The name of the icon to display on the button ([Overview of available icons](https://pictogrammers.github.io/@mdi/font/4.9.95/)).
     * `close_file`: If a file object is provided, close the file after downloading (default True).
     * `mime_type`: The mime type of the file. If not provided, the mime type will be "application/octet-stream",
           For instance setting it to "application/vnd.ms-excel" will allow the user OS to directly open the
           file into Excel.
     * `string_encoding`: The encoding to use when converting a string to bytes (default "utf-8").

    ## Note on file size

    Note that the data will be kept in memory when downloading.
    If the file is large (>10 MB), and when using [Solara server](/docs/understanding), we recommend using the
    [static files directory](/docs/reference/static-files) instead.

    """
    request_download, set_request_download = solara.use_state(False)

    # if the data changes, we 'reset'
    def reset():
        nonlocal request_download
        request_download = False
        set_request_download(False)

    solara.use_memo(reset, [data])

    # we only upload to the frontend if clicked
    def get_data() -> Optional[bytes]:
        if request_download:
            if callable(data):
                data_non_lazy = data()
            else:
                data_non_lazy = data
            if hasattr(data_non_lazy, "read"):
                if hasattr(data_non_lazy, "seek"):
                    if hasattr(data_non_lazy, "tell") and data_non_lazy.tell() != 0:
                        data_non_lazy.seek(0)
                content = data_non_lazy.read()  # type: ignore
                if close_file:
                    data_non_lazy.close()  # type: ignore
                return content
            elif isinstance(data_non_lazy, str):
                return data_non_lazy.encode(string_encoding)
            else:
                return cast(bytes, data_non_lazy)
        return None

    bytes_result: solara.Result[Optional[bytes]] = solara.use_thread(get_data, dependencies=[request_download, data])
    if filename is None and hasattr(data, "name"):
        try:
            filename = Path(data.name).name  # type: ignore
        except Exception:
            pass
    filename = filename or "solara-download.dat"
    label = label or ("Download: " + filename)
    FileDownloadWidget.element(
        filename=filename,
        bytes=bytes_result.value if bytes_result.state == solara.ResultState.FINISHED else None,
        request_download=request_download,
        on_request_download=set_request_download,
        children=children or [solara.Button(label, loading=bytes_result.state == solara.ResultState.RUNNING, icon_name=icon_name)],
        mime_type=mime_type,
    )
    if bytes_result.state == solara.ResultState.ERROR and bytes_result.error:
        raise bytes_result.error



"""# Calculator

This shows how to use `use_reducer` to implement a simple calculator.

Note that the `reducer` implements all the logic of the calculator, and the `Calculator` component is just a thin wrapper around it.



"""

import ast
import dataclasses
import operator
from typing import Any, Optional

import solara

DEBUG = False
operator_map = {
    "x": operator.mul,
    "/": operator.truediv,
    "+": operator.add,
    "-": operator.sub,
}


@dataclasses.dataclass(frozen=True)
class CalculatorState:
    input: str = ""
    output: str = ""
    left: float = 0
    right: Optional[float] = None
    operator: Any = operator.add
    error: str = ""


initial_state = CalculatorState()


def calculate(state: CalculatorState):
    result = state.operator(state.left, state.right)
    return dataclasses.replace(state, left=result)


def calculator_reducer(state: CalculatorState, action):
    action_type, payload = action
    if DEBUG:
        print("reducer", state, action_type, payload)  # noqa
    state = dataclasses.replace(state, error="")

    if action_type == "digit":
        digit = payload
        input = state.input + digit
        return dataclasses.replace(state, input=input, output=input)
    elif action_type == "percent":
        if state.input:
            try:
                value = ast.literal_eval(state.input)
            except Exception as e:
                return dataclasses.replace(state, error=str(e))
            state = dataclasses.replace(state, right=value / 100)
            state = calculate(state)
            output = f"{value / 100:,}"
            return dataclasses.replace(state, output=output, input="")
        else:
            output = f"{state.left / 100:,}"
            return dataclasses.replace(state, left=state.left / 100, output=output)
    elif action_type == "negate":
        if state.input:
            input = state.output
            input = input[1:] if input[0] == "-" else "-" + input
            output = input
            return dataclasses.replace(state, input=input, output=output)
        else:
            output = f"{-state.left:,}"
            return dataclasses.replace(state, left=-state.left, output=output)
    elif action_type == "clear":
        return dataclasses.replace(state, input="", output="")
    elif action_type == "reset":
        return initial_state
    elif action_type == "calculate":
        if state.input:
            try:
                value = ast.literal_eval(state.input)
            except Exception as e:
                return dataclasses.replace(state, error=str(e))
            state = dataclasses.replace(state, right=value)
        state = calculate(state)
        output = f"{state.left:,}"
        state = dataclasses.replace(state, output=output, input="")
        return state
    elif action_type == "operator":
        if state.input:
            state = calculator_reducer(state, ("calculate", None))
            state = dataclasses.replace(state, operator=payload, input="")
        else:
            # e.g. 2+3=*= should give 5,25
            state = dataclasses.replace(state, operator=payload, right=state.left)
        return state
    else:
        print("invalid action", action)  # noqa
        return state


@solara.component
def Calculator():
    state, dispatch = solara.use_reducer(calculator_reducer, initial_state)
    if DEBUG:
        print("->", state)  # noqa
    with solara.Card("Calculator", elevation=10, classes=["ma-4"]) as main:
        with solara.VBox(grow=False):
            solara.Text(state.error or state.output or "0")
            class_ = "pa-0 ma-1"

            with solara.HBox(grow=False):
                if state.input:
                    solara.Button("C", on_click=lambda: dispatch(("clear", None)), dark=True, class_=class_)
                else:
                    solara.Button("AC", on_click=lambda: dispatch(("reset", None)), dark=True, class_=class_)
                solara.Button("+/-", on_click=lambda: dispatch(("negate", None)), dark=True, class_=class_)
                solara.Button("%", on_click=lambda: dispatch(("percent", None)), dark=True, class_=class_)
                solara.Button("/", color="primary", on_click=lambda: dispatch(("operator", operator_map["/"])), class_=class_)

            column_op = ["x", "-", "+"]
            for i in range(3):
                with solara.HBox(grow=False):
                    for j in range(3):
                        digit = str(j + (2 - i) * 3 + 1)

                        def on_click(digit=digit):
                            dispatch(("digit", digit))

                        solara.Button(digit, on_click=on_click, class_=class_)
                    op_symbol = column_op[i]
                    op = operator_map[op_symbol]

                    def on_click_op(op=op):
                        dispatch(("operator", op))

                    solara.Button(op_symbol, color="primary", on_click=on_click_op, class_=class_)
            with solara.HBox(grow=False):

                def boom():
                    print("boom")  # noqa
                    raise ValueError("lala")

                solara.Button("?", on_click=boom, class_=class_)

                solara.Button("0", on_click=lambda: dispatch(("digit", "0")), class_=class_)
                solara.Button(".", on_click=lambda: dispatch(("digit", ".")), class_=class_)

                solara.Button("=", color="primary", on_click=lambda: dispatch(("calculate", None)), class_=class_)

    return main


Page = Calculator



from typing import Optional, Union

import reacton.ipyvuetify as v

import solara


@solara.component
def Tooltip(
    tooltip=Union[str, solara.Element],
    children=[],
    color: Optional[str] = None,
):
    """A tooltip that is shown when you hover above an element.

    Not all components support tooltips, in case it does not work,
    try wrapping the element in a `Column` or `Row`.

    ```solara
    import solara

    @solara.component
    def Page():
        with solara.Tooltip("This is a tooltip over a button"):
            solara.Button("Hover me")
        with solara.Tooltip("This is a tooltip over a text"):
            solara.Text("Hover me")
        info = solara.Info("Any component is supported as tooltip.")
        with solara.Tooltip(info, color="white"):
            with solara.Column():
                solara.Markdown("# Lorem ipsum\\n\\nDolor sit amet")
    ```

    ## Arguments

     * `tooltip`: the text, or element to display on hover.
     * `children`: the element to display the tooltip over.
     * `color`: the color of the tooltip (if None, the default color).

    """

    def set_v_on():
        for child in children:
            widget = solara.get_widget(child)
            # this only works on vue/vuetify components
            widget.v_on = "tooltip.on"  # type: ignore

    solara.use_effect(set_v_on, children)

    return v.Tooltip(
        bottom=True,
        v_slots=[
            {
                "name": "activator",
                "variable": "tooltip",
                "children": children,
            }
        ],
        color=color,
        children=[tooltip],
    )



"""# DefaultLayout

"""

import solara
from solara.website.utils import apidoc

title = "DefaultLayout"


@solara.component
def Page():
    return solara.Warning("This component does not render well as a child component. If you want to see it, you should follow the multipage guide above 🙂")


__doc__ += apidoc(solara.DefaultLayout.f)  # type: ignore



import time

from ipyvue import VueWidget


def repeat_while_false(f, timeout=10):
    while True:
        result = f()
        if result:
            return
        if timeout is not None:
            if timeout <= 0:
                raise TimeoutError("Timeout")
            timeout -= 0.1
            time.sleep(0.1)
        else:
            time.sleep(0.1)


def repeat_while_true(f, timeout=10):
    repeat_while_false(lambda: not f(), timeout)


def busy_wait_compare(getter, expected, iteration_delay=0.001, max_iterations=10_000):
    iterations = 0
    previous_value = None
    while True:
        current_value = getter()
        if current_value != previous_value:
            previous_value = current_value
            iterations = 0
        else:
            iterations += 1
        if iterations > max_iterations:
            raise TimeoutError(f"Timeout, no change in value for {max_iterations} iterations, current value: {current_value}, expected: {expected}")
        time.sleep(0.001)
        if current_value == expected:
            break


def event_and_mods(widget: VueWidget, event):
    event_match = [k for k in widget._event_handlers_map.keys() if k.startswith(event)]
    if event_match:
        return event_match[0]
    raise ValueError(f"'{event}' not found in widget {widget}")


def fire(widget: VueWidget, event):
    event_name = event_and_mods(widget, event)
    # manually call, because the CallbackDispatcher will eat exceptions
    for callback in widget._event_handlers_map[event_name].callbacks:
        callback(widget, event_name, {})


def click(widget: VueWidget, event="click"):
    fire(widget, "click")



from time import sleep

import numpy as np
from matplotlib import pyplot as plt

import solara


@solara.component
def Page():
    # define some state which will be updated regularly in a separate thread
    counter = solara.use_reactive(0)

    def render():
        """Infinite loop regularly mutating counter state"""
        while True:
            sleep(0.2)
            counter.value += 1

    # run the render loop in a separate thread
    result: solara.Result[bool] = solara.use_thread(render)
    if result.error:
        raise result.error

    # create the LiveUpdatingComponent, this component depends on the counter
    # value so will be redrawn whenever counter value changes
    LiveUpdatingComponent(counter.value)


@solara.component
def LiveUpdatingComponent(counter):
    """Component which will be redrawn whenever the counter value changes."""
    fig, ax = plt.subplots()
    ax.plot(np.arange(10), np.random.random(10))
    solara.FigureMatplotlib(fig)


Page()



from pathlib import Path
from typing import Optional

import playwright.sync_api

import solara
import solara.server.kernel_context

HERE = Path(__file__).parent


set_value = None
context: Optional["solara.server.kernel_context.VirtualKernelContext"] = None


@solara.component
def Page():
    global set_value, app_context
    value, set_value = solara.use_state(0)
    assert set_value is not None
    context = solara.server.kernel_context.get_current_context()
    assert context is not None
    solara.Text(f"Value {value}")

    def disconnect():
        assert len(context.kernel.session.websockets) == 1
        list(context.kernel.session.websockets)[0].close()

    solara.Button("Disconnect", on_click=disconnect)
    solara.Button("Increment", on_click=lambda: set_value(value + 1))

    def disconnect_and_change():
        assert len(context.kernel.session.websockets) == 1
        list(context.kernel.session.websockets)[0].close()
        set_value(100)

    solara.Button("Disconnect and change", on_click=disconnect_and_change)


def test_reconnect_simple(browser: playwright.sync_api.Browser, page_session: playwright.sync_api.Page, solara_server, solara_app, extra_include_path):
    with extra_include_path(HERE), solara_app("reconnect_test:Page"):
        page_session.goto(solara_server.base_url)
        page_session.locator("text=Value 0").wait_for()
        page_session.locator("text=Increment").click()
        page_session.locator("text=Value 1").wait_for()
        assert len(solara.server.kernel_context.contexts) == 1
        context = list(solara.server.kernel_context.contexts.values())[0]
        assert len(context.kernel.session.websockets) == 1
        ws = list(context.kernel.session.websockets)[0]
        page_session.locator("text=Disconnect").nth(0).click()
        n = 0
        # we wait till the current websocket is not connected anymore, and a different one is connected
        while not (ws not in context.kernel.session.websockets and len(context.kernel.session.websockets) == 1):
            page_session.wait_for_timeout(100)
            n += 1
            if n > 50:
                raise RuntimeError("Timeout waiting for reconnected websocket")
        page_session.locator("text=Value 1").wait_for()
        page_session.locator("text=Increment").click()
        page_session.locator("text=Value 2").wait_for()
        # we should not have created a new context
        assert len(solara.server.kernel_context.contexts) == 1
        page_session.goto("about:blank")
        assert context.closed_event.wait(10)


def test_reconnect_fail(browser: playwright.sync_api.Browser, page_session: playwright.sync_api.Page, solara_server, solara_app, extra_include_path):
    with extra_include_path(HERE), solara_app("reconnect_test:Page"):
        # import reconnect_test as module

        page_session.goto(solara_server.base_url)
        page_session.locator("text=Value 0").wait_for()
        page_session.locator("text=Increment").click()
        page_session.locator("text=Value 1").wait_for()
        cull_timeout_previous = solara.server.settings.kernel.cull_timeout
        context = None
        try:
            solara.server.settings.kernel.cull_timeout = "0s"
            assert len(solara.server.kernel_context.contexts) == 1
            context = list(solara.server.kernel_context.contexts.values())[0]
            assert len(context.kernel.session.websockets) == 1
            page_session.locator("text=Disconnect").nth(0).click()
            page_session.locator("text=Could not restore session").wait_for()
            n = 0
            # we wait till the all contexts are closed
            while len(solara.server.kernel_context.contexts):
                page_session.wait_for_timeout(100)
                n += 1
                if n > 50:
                    raise RuntimeError("Timeout waiting for kernel shutdown")

        finally:
            page_session.goto("about:blank")
            if context is not None:
                assert context.closed_event.wait(10)
            solara.server.settings.kernel.cull_timeout = cull_timeout_previous


def test_reconnect_and_update(browser: playwright.sync_api.Browser, page_session: playwright.sync_api.Page, solara_server, solara_app, extra_include_path):
    with extra_include_path(HERE), solara_app("reconnect_test:Page"):
        page_session.goto(solara_server.base_url)
        page_session.locator("text=Value 0").wait_for()
        page_session.locator("text=Increment").click()
        page_session.locator("text=Value 1").wait_for()

        assert len(solara.server.kernel_context.contexts) == 1
        context = list(solara.server.kernel_context.contexts.values())[0]

        # this will mutate when disconnected, testing the update functionality on reconnect
        page_session.locator("text=Disconnect and change").click()
        page_session.locator("text=Value 100").wait_for()
        page_session.goto("about:blank")
        assert context.closed_event.wait(10)



import solara
from solara.alias import rv


@solara.component
def Details(summary="Summary", children=[], expand=False):
    expand, set_expand = solara.use_state_or_update(expand)

    def on_v_model(v_model):
        if v_model is None:  # collapsed:
            set_expand(False)
        elif v_model == 0:
            set_expand(True)
        else:
            raise RuntimeError(f"v_model has odd value: {v_model}")

    with rv.ExpansionPanels(v_model=0 if expand else None, on_v_model=on_v_model) as main:
        with rv.ExpansionPanel():
            rv.ExpansionPanelHeader(children=[summary])
            rv.ExpansionPanelContent(children=children)
    return main



# this module will be conditionally imported in patch.py for now
# in the future, we may want to move the esm features of ipyreact
# into a separate package, and then we can import it unconditionally
import logging
from pathlib import Path
from typing import Dict, List, Tuple, Union

import ipyreact.importmap
import ipyreact.module

logger = logging.getLogger("solara.server.esm")

_modules: Dict[str, Tuple[Union[str, Path], List[str]]] = {}


# in solara server, we'll monkey patch ipyreact.module with this
def define_module(name, module: Union[str, Path]):
    # collect the dependencies at this moment
    dependencies = list(_modules.keys())
    logger.info("define module %s %s (dependencies=%r)", name, module, dependencies)
    if name in _modules:
        old_module, dependencies = _modules[name]
    _modules[name] = (module, dependencies)
    return None


def get_module_names():
    return list(_modules.keys())


def create_modules():
    logger.info("define modules %s", _modules)
    widgets = {}
    for name, (module, dependencies) in _modules.items():
        widgets[name] = create_module(name, module, dependencies=dependencies)
    return widgets


def create_module(name, module: Union[str, Path], dependencies: List[str]):
    return ipyreact.module.Module(code=module if not isinstance(module, Path) else module.read_text(encoding="utf8"), name=name, dependencies=dependencies)


def create_import_map():
    logger.info("create import map %s", ipyreact.importmap._effective_import_map)
    return ipyreact.importmap.ImportMap(import_map=ipyreact.importmap._effective_import_map)



import os
import re
import site
import sys
import uuid
import warnings
from enum import Enum
from pathlib import Path
from typing import Optional

from filelock import FileLock

import solara.util
from solara.minisettings import BaseSettings

from .. import (  # noqa  # sidefx is that this module creates the ~/.solara directory
    settings,
)
from ..util import get_solara_home

if site.getuserbase() and __file__.startswith(site.getuserbase()):
    prefix = site.getuserbase()
else:
    prefix = sys.prefix


class ThemeVariant(str, Enum):
    light = "light"
    dark = "dark"
    auto = "auto"


class ThemeSettings(BaseSettings):
    variant: ThemeVariant = ThemeVariant.light
    loader: str = "solara"

    class Config:
        env_prefix = "solara_theme_"
        case_sensitive = False
        env_file = ".env"


class SSG(BaseSettings):
    # the first app create will initialize this if it is not set
    build_path: Optional[Path] = None
    enabled: bool = False
    headed: bool = False

    class Config:
        env_prefix = "solara_ssg_"
        case_sensitive = False
        env_file = ".env"


class Search(BaseSettings):
    enabled: bool = False

    class Config:
        env_prefix = "solara_search_"


class Telemetry(BaseSettings):
    mixpanel_token: str = "91845eb13a68e3db4e58d64ad23673b7"
    mixpanel_enable: bool = True
    server_user_id: str = "not_set"
    server_fingerprint: str = str(uuid.getnode())
    server_session_id: str = str(uuid.uuid4())

    class Config:
        env_prefix = "solara_telemetry_"
        case_sensitive = False
        env_file = ".env"


class Assets(BaseSettings):
    proxy_cache_dir: Path = Path(prefix + "/share/solara/cdn/")
    fontawesome_enabled: bool = True
    fontawesome_path: str = "/font-awesome@4.5.0/css/font-awesome.min.css"

    class Config:
        env_prefix = "solara_assets_"
        case_sensitive = False
        env_file = ".env"


class Kernel(BaseSettings):
    cull_timeout: str = "24h"
    max_count: Optional[int] = None

    class Config:
        env_prefix = "solara_kernel_"
        case_sensitive = False
        env_file = ".env"


AUTH0_TEST_CLIENT_ID = "cW7owP5Q52YHMZAnJwT8FPlH2ZKvvL3U"
AUTH0_TEST_CLIENT_SECRET = "zxITXxoz54OjuSmdn-PluQgAwbeYyoB7ALlnLoodftvAn81usDXW0quchvoNvUYD"
AUTH0_TEST_API_BASE_URL = "dev-y02f2bpr8skxu785.us.auth0.com"
AUTH0_LOGOUT_PATH = "v2/logout"

FIEF_TEST_CLIENT_ID = "x2np62qgwp6hnEGTP4JYUE3igdZWhT-AvjpjwwDyKXU"
FIEF_TEST_CLIENT_SECRET = "XQlByE1pVIz5h2SBN2GYDwT_ziqArHJgLD3KqMlCHjg"
FIEF_TEST_API_BASE_URL = "solara-dev.fief.dev"
FIEF_LOGOUT_PATH = "logout"
SESSION_SECRET_KEY_DEFAULT = "change me"

OAUTH_TEST_CLIENT_IDs = [AUTH0_TEST_CLIENT_ID, FIEF_TEST_CLIENT_ID]


class Session(BaseSettings):
    secret_key: str = SESSION_SECRET_KEY_DEFAULT
    https_only: Optional[bool] = None
    same_site: str = "lax"

    class Config:
        env_prefix = "solara_session_"
        case_sensitive = False
        env_file = ".env"


class OAuth(BaseSettings):
    private: bool = False

    client_id: str = AUTH0_TEST_CLIENT_ID
    client_secret: str = AUTH0_TEST_CLIENT_SECRET
    api_base_url: str = AUTH0_TEST_API_BASE_URL
    logout_path: str = AUTH0_LOGOUT_PATH
    scope: str = "openid profile email"

    class Config:
        env_prefix = "solara_oauth_"
        case_sensitive = False
        env_file = ".env"


HOST_DEFAULT = os.environ.get("HOST", "localhost")
is_mac_os_conda = "arm64-apple-darwin" in HOST_DEFAULT
is_wsl_windows = re.match(r".*?-w1[0-9]", HOST_DEFAULT)
if is_mac_os_conda or is_wsl_windows:
    HOST_DEFAULT = "localhost"


class MainSettings(BaseSettings):
    use_pdb: bool = False
    mode: str = "production"
    tracer: bool = False
    timing: bool = False
    root_path: Optional[str] = None  # e.g. /myapp (without trailing slash)
    base_url: str = ""  # e.g. https://myapp.solara.run/myapp/
    platform: str = sys.platform
    host: str = HOST_DEFAULT
    experimental_performance: bool = False

    class Config:
        env_prefix = "solara_"
        case_sensitive = False
        env_file = ".env"


main = MainSettings()
theme = ThemeSettings()
telemetry = Telemetry()
ssg = SSG()
search = Search()
assets = Assets()
oauth = OAuth()
session = Session()
kernel = Kernel()
# fail early
solara.util.parse_timedelta(kernel.cull_timeout)

if settings.assets.proxy:
    try:
        assets.proxy_cache_dir.mkdir(exist_ok=True, parents=True)
    except OSError as e:
        settings.assets.proxy = False
        warnings.warn(
            f"Could not create {assets.proxy_cache_dir} due to {e}. We will automatically disable the assets proxy for you. "
            "If you want to disable this warning, set SOLARA_ASSETS_PROXY to False (e.g. export SOLARA_ASSETS_PROXY=false). "
            "Or change the SOLARA_PROXY_CACHE_DIR environment variable to a directory where you have write access."
        )
        # that's ok, the user probably doesn't have permission to create the directory
        # in this case, we would need to install solara-assets?
        pass

if telemetry.server_user_id == "not_set":
    home = get_solara_home()
    server_user_id_file = home / "server_user_id.txt"
    try:
        with FileLock(str(server_user_id_file) + ".lock"):
            if not server_user_id_file.exists():
                server_user_id_file.write_text(str(uuid.uuid4()))
            telemetry.server_user_id = server_user_id_file.read_text()
    except OSError:
        pass  # it's ok

if oauth.client_id:
    if oauth.client_id not in OAUTH_TEST_CLIENT_IDs:
        if session.secret_key == SESSION_SECRET_KEY_DEFAULT:
            raise ValueError(
                "You must set a session secret key for oauth, it is not safe to use the default. Please set SOLARA_SESSION_SECRET_KEY in your environment."
            )
        if session.https_only is None:
            raise ValueError(
                "You must set https_only for the session to True (recommended) or False in your are using your own oauth provider."
                "Please set SOLARA_SESSION_HTTPS_ONLY in your environment to True or False"
            )
    else:
        # for the test accounts, this is fine
        if session.https_only is None:
            session.https_only = False



import pytest
from playwright.sync_api import Page, expect

# Tests are only run in solara, since others don't support parametrization.


# Tests rendering of menus, as well as click functionality on child element
@pytest.mark.parametrize("test_type", ["Menu", "ClickMenu", "ContextMenu"])
def test_menus_playwright(test_type, solara_test, page_session: Page, assert_solara_snapshot):
    import solara
    from solara.lab import ClickMenu, ContextMenu, Menu

    @solara.component
    def Page():
        # without height dropdown menu won't be visible in screenshots
        with solara.Column(classes=["test-class-container"], style="height:300px; width: 400px;"):
            text = solara.use_reactive("pre-test")

            def on_click():
                text.set("Works!")

            solara.Text(text.value, classes=["test-class-text"])

            menu_activator = solara.Button("activator")
            menu_child = solara.Button("child", on_click=on_click, classes=["test-class-child"])

            if test_type == "Menu":
                Menu(activator=menu_activator, children=[menu_child])
            elif test_type == "ClickMenu":
                ClickMenu(activator=menu_activator, children=[menu_child])
            else:
                ContextMenu(activator=menu_activator, children=[menu_child])

    solara.display(Page())

    text_el = page_session.locator(".test-class-text")
    expect(text_el).to_contain_text("pre-test")
    activator_button = page_session.locator("text=activator")
    expect(page_session.locator(".test-class-child")).to_have_count(0)
    if test_type == "ContextMenu":
        activator_button.click(button="right")
    else:
        activator_button.click()

    child_button = page_session.locator("text=child")
    page_session.wait_for_timeout(350)  # Wait for any animations after click
    expect(page_session.locator(".test-class-child")).to_be_visible()
    child_button.click()
    page_session.wait_for_timeout(350)
    expect(text_el).to_contain_text("Works!")


# Tests successive click behaviour of menus
@pytest.mark.parametrize("test_type", ["Menu", "ClickMenu", "ContextMenu"])
def test_menus_successive(test_type, solara_test, page_session: Page):
    import solara
    from solara.lab import ClickMenu, ContextMenu, Menu

    @solara.component
    def Page():
        menu_activator = solara.Div(style="height: 200px; width: 400px;", classes=["test-class-activator"])
        menu_child = solara.Div(classes=["test-class-menu"], style="height: 100px; width: 50px;")

        if test_type == "Menu":
            Menu(activator=menu_activator, children=[menu_child])
        elif test_type == "ClickMenu":
            ClickMenu(activator=menu_activator, children=[menu_child])
        else:
            ContextMenu(activator=menu_activator, children=[menu_child])

    solara.display(Page())

    activator_el = page_session.locator(".test-class-activator")
    expect(page_session.locator(".test-class-menu")).to_have_count(0)  # Menu should not exist yet
    if test_type == "ContextMenu":
        activator_el.click(button="right")
    else:
        activator_el.click()

    page_session.wait_for_timeout(350)  # Wait for any animations after click
    expect(page_session.locator(".test-class-menu")).to_be_in_viewport()  # Menu should be visible now

    if test_type == "ContextMenu":
        activator_el.click(button="right")
        page_session.wait_for_timeout(350)
        expect(page_session.locator(".test-class-menu")).to_be_in_viewport()
    else:
        activator_el.click()
        page_session.wait_for_timeout(350)
        expect(page_session.locator(".test-class-menu")).not_to_be_in_viewport()  # After second click, Menu and ClickMenu should disappear



from matplotlib.figure import Figure

import solara
import solara.server.patch
from solara.server import kernel


@solara.component
def Page():
    # do this instead of plt.figure()
    fig = Figure()
    ax = fig.subplots()
    ax.plot([1, 2, 3], [1, 4, 9])
    return solara.FigureMatplotlib(fig)


def test_render():
    box, rc = solara.render(Page(), handle_error=False)
    assert len(box.children) == 1


def test_pylab(no_kernel_context):
    cleanup = solara.server.patch.patch_matplotlib()
    try:
        kernel_1 = kernel.Kernel()
        context_1 = solara.server.kernel_context.VirtualKernelContext(id="1", kernel=kernel_1, session_id="session-1")
        kernel_2 = kernel.Kernel()
        context_2 = solara.server.kernel_context.VirtualKernelContext(id="2", kernel=kernel_2, session_id="session-1")
        import matplotlib.pyplot as plt
        from matplotlib._pylab_helpers import Gcf

        assert len(Gcf.get_all_fig_managers()) == 0
        plt.figure()

        assert len(Gcf.get_all_fig_managers()) == 1

        default_color = (1, 1, 1, 0)
        white = "white"
        black = "black"
        assert plt.rcParams["figure.facecolor"] in [default_color, white]
        plt.style.use("default")
        assert plt.rcParams["figure.facecolor"] == white

        plt.style.use("dark_background")
        assert plt.rcParams["figure.facecolor"] == black

        with context_1:
            assert len(Gcf.get_all_fig_managers()) == 0
            plt.figure()
            assert plt.rcParams["figure.facecolor"] == black
            assert len(Gcf.get_all_fig_managers()) == 1
            plt.style.use("default")
            assert plt.rcParams["figure.facecolor"] == white

        assert plt.rcParams["figure.facecolor"] == black
        assert len(Gcf.get_all_fig_managers()) == 1
        plt.style.use("default")
        assert plt.rcParams["figure.facecolor"] == white

        with context_2:
            assert plt.rcParams["figure.facecolor"] == white
            assert len(Gcf.get_all_fig_managers()) == 0
            plt.figure()
            assert len(Gcf.get_all_fig_managers()) == 1
            plt.figure()
            assert len(Gcf.get_all_fig_managers()) == 2
            plt.style.use("dark_background")
            assert plt.rcParams["figure.facecolor"] == black
        with context_1:
            assert len(Gcf.get_all_fig_managers()) == 1
            assert plt.rcParams["figure.facecolor"] == white
        assert plt.rcParams["figure.facecolor"] == white

    finally:
        cleanup()
        context_1.close()
        context_2.close()



# import inspect
# import urllib.parse

import solara

title = "Examples"


@solara.component
def Page():
    # show a gallery of all the examples
    router = solara.use_router()
    route_current = router.path_routes[-2]

    for route in route_current.children:
        if route.children:
            solara.Markdown(f"## {route.label}\n" + (route.module.__doc__ or ""))
            with solara.ColumnsResponsive(12, 6, 6, 6, 4):
                for child in route.children:
                    path = route.path + "/" + child.path
                    if child.path in [
                        "tokenizer",
                        "sine",
                        "authorization",
                        "layout_demo",
                        "multipage",
                        "scatter",
                        "scrolling",
                        "tutorial_streamlit",
                        "login_oauth",
                        "pokemon_search",
                        "altair",
                        "bqplot",
                        "ipyleaflet",
                        "calculator",
                        "countdown_timer",
                        "todo",
                    ]:
                        image = route.path + "/" + child.path + ".png"
                        image_url = "https://dxhl76zpt6fap.cloudfront.net/public/examples/" + image
                    else:
                        image_url = "https://dxhl76zpt6fap.cloudfront.net/public/logo.svg"

                    path = getattr(child.module, "redirect", path)
                    if path:
                        with solara.Card(child.label, style="height: 100%;"):
                            with solara.Link(path):
                                with solara.Column(align="center"):
                                    solara.Image(image_url, width="120px" if image_url.endswith(".svg") else "100%")


@solara.component
def Layout(children):
    # TODO: this is using a private API, what is the best way to do this?
    # we want to 'eat' the whole route for the current level, and the level below
    # for example the utilities directory. But if an example does routing, we don't want
    # to take on that route.
    router = solara.use_router()
    route_current = router.path_routes[-1]
    # route_current, all_routes = solara.use_route()

    if route_current is None:
        return solara.Error("Page not found")
    module = route_current.module
    assert module is not None
    github_url = solara.util.github_url(module.__file__)

    with solara.HBox(grow=False) as main:
        if route_current.path == "fullscreen":
            with solara.Padding(4, children=children):
                pass
        else:
            with solara.VBox(grow=True, align_items="baseline"):
                doc = module.__doc__
                if doc:
                    with solara.VBox(grow=True):
                        solara.Markdown(doc)
                with solara.HBox():
                    if route_current.path != "/":
                        solara.Button("View source code on GitHub", icon_name="mdi-github-circle", href=github_url, class_="ma-2", target="_blank", text=True)
                        # code = inspect.getsource(module)

                        # code_quoted = urllib.parse.quote_plus(code)
                        # url = f"https://test.solara.dev/try?code={code_quoted}"
                        # solara.Button("Run on solara.dev", icon_name="mdi-pencil", href=url, class_="ma-2", target="_blank")
                # with solara.HBox():
                if not hasattr(module, "Page"):
                    solara.Error(f"No Page component found in {module}")
                else:
                    with solara.Padding(4, children=children):
                        pass
    return main



"""# Markdown
"""


import solara
from solara.alias import rv
from solara.website.utils import apidoc


@solara.component
def Page():
    markdown_initial = """
# Large
## Smaller

## List items

    * item 1
    * item 2


## Code highlight support
```python
code = "formatted" and "supports highlighting"
```


## Mermaid support!
See [Mermaid docs](https://mermaid-js.github.io/)

```mermaid
graph TD;
    A-->B;
    A-->C;
    B-->D;
    C-->D;
```


    """.strip()

    markdown_text, set_markdown_text = solara.use_state(markdown_initial)
    # with solara.GridFixed(columns=2) as main:
    with solara.HBox(grow=True) as main:
        with solara.VBox():
            solara.Markdown("# Input text")
            with solara.Padding(2):
                with rv.Sheet(elevation=2):
                    rv.Textarea(v_model=markdown_text, on_v_model=set_markdown_text, rows=30)
        with solara.VBox():
            solara.Markdown("# Renders like")
            with solara.Padding(2):
                with rv.Sheet(elevation=2):
                    solara.Markdown(markdown_text)

    return main


__doc__ += apidoc(solara.Markdown.f)  # type: ignore



import inspect
import textwrap


def sig(f):
    lines = inspect.getsourcelines(f)[0]
    end = [k.endswith(":\n") for k in lines].index(True)
    lines = lines[: end + 1]
    return "".join(lines)


def code(f):
    lines = inspect.getsourcelines(f)[0]
    def_end = [k.endswith(":\n") for k in lines].index(True)
    doc_lines = len(f.__doc__.split("\n"))
    lines = lines[: def_end + 1] + lines[def_end + 1 + doc_lines :]

    return "".join(lines)


def apidoc(f, full=False):
    if not f.__doc__:
        return "no docstring"

    doclines = f.__doc__.split("\n")
    first = doclines[0].strip()
    rest = "\n".join(doclines[1:])  # .strip()
    if full:
        return f"""
{first}

```python
{code(f)}
    ...
```


{textwrap.dedent(rest)}
"""
    else:
        return f"""
{first}

```python
{sig(f)}
    ...
```


{textwrap.dedent(rest)}
"""



"""#Memoize

"""
import time

import solara
import solara.lab
from solara.website.utils import apidoc

# make sure the cache is only created once
storage = solara.cache.Memory(max_items=2)


@solara.component
def Page():
    x, set_x = solara.use_state(5)

    @solara.memoize(storage=storage)
    def long_running_function(x: int) -> int:
        """This function takes a long time to run."""
        time.sleep(3)
        return x**2

    result = long_running_function.use_thread(x)

    with solara.Card("Expensive computation") as main:
        solara.Markdown("We cache 2 values. Each computation takes 3 seconds. If you go back to a cached value, you will see the result immediately.")
        solara.IntSlider("x", value=x, on_value=set_x)
        if result.state == solara.ResultState.FINISHED:
            solara.Markdown(f"Square of {x} is {result.value}")
        else:
            solara.Markdown("Running...")
    return main


__doc__ += apidoc(solara.memoize)  # type: ignore



"""
# get_session_id

"""
import solara
from solara.website.utils import apidoc

from . import NoPage

title = "get_session_id"


Page = NoPage


__doc__ += apidoc(solara.get_session_id)  # type: ignore



"""
# ipyleaflet advanced

Extends the [basic ipyleaflet example](/examples/libraries/ipyleaflet) with a marker that can be dragged around, and a
dropdown to select the map style. Two buttons allow to reset the map to the default zoom and center and to zoom
to the marker.
"""
import ipyleaflet

import solara

center_default = (53.2305799, 6.5323552)
zoom_default = 5
maps = {
    "OpenStreetMap.Mapnik": ipyleaflet.basemaps.OpenStreetMap.Mapnik,
    "OpenTopoMap": ipyleaflet.basemaps.OpenTopoMap,
    "Esri.WorldTopoMap": ipyleaflet.basemaps.Esri.WorldTopoMap,
}

zoom = solara.reactive(zoom_default)
center = solara.reactive(center_default)
marker_location = solara.reactive(center_default)

map_name = solara.reactive(list(maps)[0])


@solara.component
def Page():
    def location_changed(location):
        # do things with the location
        marker_location.set(location)

    with solara.Column(style={"min-width": "500px", "height": "500px"}):
        solara.Markdown(f"Market set to: {marker_location.value}", style={"color": "#6e6e6e"})

        map = maps[map_name.value]
        url = map.build_url()

        def goto_marker():
            center.value = marker_location.value
            zoom.value = 13

        def reset_view():
            center.value = center_default
            zoom.value = zoom_default

        solara.Select(label="Map", value=map_name, values=list(maps))
        solara.SliderInt(label="Zoom level", value=zoom, min=1, max=20)
        with solara.Row():
            solara.Button(label="Zoom to marker", on_click=goto_marker)
            solara.Button(label="Reset view", on_click=reset_view)

        # Isolation is required to prevent the map from overlapping navigation (when screen width < 960px)
        with solara.Column(style={"isolation": "isolate"}):
            ipyleaflet.Map.element(  # type: ignore
                zoom=zoom.value,
                on_zoom=zoom.set,
                center=center.value,
                on_center=center.set,
                scroll_wheel_zoom=True,
                layers=[
                    ipyleaflet.TileLayer.element(url=url),
                    ipyleaflet.Marker.element(location=marker_location.value, draggable=True, on_location=location_changed),
                ],
            )



from pathlib import Path

from solara.autorouting import generate_routes_directory

title = "Docs"
HERE = Path(__file__)
# if we didn't put the content in the subdirectory, but pointed to the current file
# we would include the current file recursively, causing an infinite loop
routes = generate_routes_directory(HERE.parent / "content")



import solara
from solara.alias import rv

from ..components import article, data
from ..data import articles, names


@solara.component
def PeopleCard(name):
    with solara.Card(f"Employee of the Month: {name}") as main:
        with rv.CardText():
            solara.Markdown(
                """
                * Department: foo
                * Skills: bar, baz
                """
            )
            with solara.Link(f"/people/{name}"):
                solara.Button("View employee", text=True, icon_name="mdi-profile")
    return main


@solara.component
def Layout(children=[]):
    router = solara.use_context(solara.routing.router_context)
    with solara.VBox() as navigation:
        with rv.List(dense=True):
            with rv.ListItemGroup(v_model=router.path):
                with solara.Link(solara.resolve_path("/")):
                    with solara.ListItem("Home", icon_name="mdi-home", value="/"):
                        pass
                with solara.ListItem("tabular data", icon_name="mdi-database"):
                    for name in names:
                        pathname = f"/tabular/{name}"
                        with solara.Link(solara.resolve_path(pathname)):
                            solara.ListItem(name, value=pathname)
                with solara.ListItem("Articles", icon_name="mdi-book-open"):
                    for name, article_ in articles.items():
                        pathname = f"/article/{name}"
                        with solara.Link(solara.resolve_path(pathname)):
                            solara.ListItem(article_.title, value=pathname)

    with solara.AppLayout(navigation=navigation, title="Solara demo", children=children) as main:
        pass
    return main


@solara.component
def Page():
    with solara.VBox() as main:
        solara.Title("Solara demo » Home")
        data.Overview()
        article.Overview()

        with solara.ColumnsResponsive(12):
            with solara.Card("Other"):
                with solara.ColumnsResponsive(6):
                    PeopleCard("Maarten Breddels")
                    with solara.Card("Quick links"):
                        with solara.Column():
                            for name in names:
                                with solara.Link(f"/viz/scatter/{name}"):
                                    solara.Button(f"Scatter for {name}", text=True)
                                with solara.Link(f"/viz/histogram/{name}"):
                                    solara.Button(f"Histogram for {name}", text=True)

    return main



import solara

from . import user

github_url = solara.util.github_url(__file__)


@solara.component
def Page():
    assert user.value is not None
    solara.Markdown(f"Hi {user.value.username}!")
    solara.Button(label="View source", icon_name="mdi-github-circle", attributes={"href": github_url, "target": "_blank"}, text=True, outlined=True)



"""
# Style

"""
import solara
from solara.website.utils import apidoc


@solara.component
def Page():
    insert_css, set_insert_css = solara.use_state(True)

    css = """
    .mybutton {
        font-family: Serif;
    }

    /* this selector has to be very specific to override the vuetify style */
    .v-btn.mybutton {
        color: #4CAF50; /* Green */
    }
    /* vuetify's background color css has very high CSS-specificity, so we use !important */
    .mybutton {
        background-color: #FF9800 !important; /* Orange */
    }
    """

    with solara.VBox() as main:
        solara.Checkbox(label="Use CSS", value=insert_css, on_value=set_insert_css)
        solara.Markdown(
            f"""
## CSS Example that styles the button below
```css
{css}
```
"""
        )
        if insert_css:
            solara.Style(css)
        solara.Button(label="Advanced users might want to style this", icon_name="mdi-thumb-up", classes=["mybutton"])
    return main


__doc__ += apidoc(solara.Style.f)  # type: ignore



from typing import Callable

import solara
import solara.lab
from solara.alias import rv
from solara.server import settings


@solara._component_vue("algolia.vue")
def Algolia(app_id: str, index_name: str, api_key: str, debug=False):
    pass


@solara.component
def Header(
    on_toggle_left_menu: Callable[[], None] = None,
    on_toggle_right_menu: Callable[[], None] = None,
):
    # use routes of parent (assuming we are a child of a layout)
    route_current, all_routes = solara.use_route(level=-1)
    router = solara.use_router()
    dark_effective = solara.lab.use_dark_effective()

    # set states for menu
    with solara.Column(gap="0px"):
        with solara.Div(classes=["news"]):
            solara.HTML(
                "div",
                unsafe_innerHTML="<a href='https://github.com/widgetti/solara' target='_blank' >Star us on github 🤩</a>",
            )
        with rv.AppBar(tag="header", flat=True, class_="bg-primary-fade padding-40", height="auto"):
            with rv.ToolbarTitle(class_="d-flex", style_="align-items:center"):
                if route_current and len(route_current.children) > 0:
                    with solara.Button(icon=True, class_="hidden-md-and-up", on_click=lambda: on_toggle_left_menu and on_toggle_left_menu()):
                        rv.Icon(children=["mdi-menu"])
                with solara.Link(path_or_route="/"):
                    solara.Image(router.root_path + f"/static/assets/images/logo{'_white' if dark_effective else ''}.svg")
            rv.Spacer()

            if settings.search.enabled:
                from solara_enterprise.search.search import Search

                Search()
            Algolia(app_id="9KW9L7O5EQ", api_key="ef7495102afff1e16d1b7cf6ec2ab2d0", index_name="solara", debug=True)
            # menu
            with rv.Html(tag="ul", class_="main-menu menu d-none d-md-flex"):
                for route in all_routes:
                    if route.path == "apps":
                        continue
                    current = route_current == route
                    with rv.Html(tag="li", class_="active" if current else None):
                        solara.Link("/" + route.path if route.path != "/" else "/", children=[route.label])
            with rv.Btn(icon=True, tag="a", class_="d-none d-md-flex", attributes={"href": solara.github_url, "target": "_blank"}):
                rv.Icon(children=["mdi-github-circle"])

            with rv.Btn(icon=True, tag="a", class_="d-none d-md-flex", attributes={"href": "https://discord.gg/dm4GKNDjXN", "target": "_blank"}):
                rv.Icon(children=["mdi-discord"])

            solara.lab.ThemeToggle()

            with solara.Button(icon=True, class_="hidden-md-and-up", on_click=lambda: on_toggle_right_menu and on_toggle_right_menu()):
                rv.Icon(children=["mdi-menu"])



import solara

from ... import data
from ...components import Layout


@solara.component
def Page(name: str, page: int = 0, page_size=100):
    if name not in data.dfs:
        return solara.Error(f"No such dataframe: {name!r}")
    df = data.dfs[name]
    with Layout() as main:
        solara.DataTable(df=df, page=page, items_per_page=page_size)
    return main



import contextlib
import hashlib
import json
import logging
import os
import sys
import time
from pathlib import Path
from typing import Dict, List, Optional, Tuple, TypeVar

import ipykernel
import ipyvue
import ipywidgets
import jinja2
import requests

import solara
import solara.routing
import solara.settings
from solara.lab import cookies as solara_cookies
from solara.lab import headers as solara_headers

from . import app, jupytertools, patch, settings, websocket
from .kernel import Kernel, deserialize_binary_message
from .kernel_context import initialize_virtual_kernel

COOKIE_KEY_SESSION_ID = "solara-session-id"

T = TypeVar("T")


directory = Path(__file__).parent
template_name = "index.html.j2"
ipykernel_major = int(ipykernel.__version__.split(".")[0])
ipywidgets_major = int(ipywidgets.__version__.split(".")[0])
cache_memory = solara.cache.Memory(max_items=128)
vue3 = ipyvue.__version__.startswith("3")

# first look at the project directory, then the builtin solara directory


@solara.memoize(storage=cache_memory)
def get_jinja_env(app_name: str) -> jinja2.Environment:
    jinja_loader = jinja2.FileSystemLoader(
        [
            app.apps["__default__"].directory.parent / "templates",
            str(directory / "templates"),
        ]
    )
    jinja_env = jinja2.Environment(loader=jinja_loader, autoescape=True)
    return jinja_env


logger = logging.getLogger("solara.server.server")
nbextensions_ignorelist = [
    "jupytext/index",
    "nbextensions_configurator/config_menu/main",
    "jupytext/index",
    "nbdime/index",
    "voila/extension",
    "contrib_nbextensions_help_item/main",
    "execute_time/ExecuteTime",
    "dominocode/extension",
    "low-code-assistant/extension",
    "domino-code-assist/extension",
    "jupyter-js/extension",
    "jupyter-js-widgets/extension",
    "jupyter_dash/main",
]


def readyz():
    return {"status": "ok"}, 200


def wait_ready(url, timeout=10) -> None:
    """Wait for a solara server at root url to be ready, or throw a TimeoutError

    This uses the /readyz endpoint to check if the server is ready.

    >>> solara.server.server.wait_ready("http://localhost:8888")
    ...

    """
    t0 = time.time()
    while True:
        try:
            r = requests.get(url + "/readyz")
            if r.status_code == 200:
                return
        except Exception:
            pass
        time.sleep(0.1)
        if time.time() - t0 > timeout:
            raise TimeoutError(f"Timeout waiting for {url}")


def is_ready(url) -> bool:
    """Returns whether a solara server at root url is ready.

    This uses the /readyz endpoint to check if the server is ready.

    >>> solara.server.server.is_ready("http://localhost:8888")
    True

    """
    try:
        r = requests.get(url + "/readyz")
        if r.status_code == 200:
            return True
    except Exception:
        pass
    return False


async def app_loop(ws: websocket.WebsocketWrapper, cookies: Dict[str, str], headers, session_id: str, kernel_id: str, page_id: str, user: dict = None):
    context = initialize_virtual_kernel(session_id, kernel_id, ws)
    if context is None:
        logging.warning("invalid kernel id: %r", kernel_id)
        # to avoid very fast reconnects (we are in a thread anyway)
        time.sleep(0.5)
        return

    if settings.main.tracer:
        import viztracer

        output_file = f"viztracer-{page_id}.html"
        run_context = viztracer.VizTracer(output_file=output_file, max_stack_depth=10)
        logger.warning(f"Running with tracer: {output_file}")
    else:
        run_context = solara.util.nullcontext()

    kernel = context.kernel
    try:
        context.page_connect(page_id)
        with run_context, context:
            if user:
                from solara_enterprise.auth import user as solara_user

                solara_user.set(user)

            solara_cookies.set(cookies)  # type: ignore
            solara_headers.set(headers)  # type: ignore

            while True:
                if settings.main.timing:
                    widgets_ids = set(patch.widgets)
                try:
                    message = await ws.receive()
                except websocket.WebSocketDisconnect:
                    try:
                        context.kernel.session.websockets.remove(ws)
                    except KeyError:
                        pass
                    logger.debug("Disconnected")
                    break
                t0 = time.time()
                if isinstance(message, str):
                    msg = json.loads(message)
                else:
                    msg = deserialize_binary_message(message)
                t1 = time.time()
                if not process_kernel_messages(kernel, msg):
                    # if we shut down the kernel, we do not keep the page session alive
                    context.close()
                    return
                t2 = time.time()
                if settings.main.timing:
                    widgets_ids_after = set(patch.widgets)
                    created_widgets_count = len(widgets_ids_after - widgets_ids)
                    close_widgets_count = len(widgets_ids - widgets_ids_after)
                    print(  # noqa: T201
                        f"timing: total={t2-t0:.3f}s, deserialize={t1-t0:.3f}s, kernel={t2-t1:.3f}s"
                        f" widget: created: {created_widgets_count} closed: {close_widgets_count}"
                    )
    finally:
        context.page_disconnect(page_id)


def process_kernel_messages(kernel: Kernel, msg: Dict) -> bool:
    session = kernel.session
    assert session is not None
    comm_manager = kernel.comm_manager

    def send_status(status, parent):
        session.send(
            kernel.iopub_socket,
            "status",
            {"execution_state": status},
            parent=parent,
            ident=None,
        )

    @contextlib.contextmanager
    def busy_idle(parent):
        send_status("busy", parent=parent)
        try:
            yield
        finally:
            send_status("idle", parent=parent)

    msg_type = msg["header"]["msg_type"]

    if ipykernel_major < 6:
        # the channel argument was added in 6.0
        kernel.set_parent(None, msg)
    else:
        kernel.set_parent(None, msg, msg["channel"])
    if msg_type == "kernel_info_request":
        content = {
            "status": "ok",
            "protocol_version": "5.3",
            "implementation": Kernel.implementation,
            "implementation_version": Kernel.implementation_version,
            "language_info": {},
            "banner": Kernel.banner,
            "help_links": [],
        }
        with busy_idle(msg["header"]):
            msg = kernel.session.send(kernel.shell_stream, "kernel_info_reply", content, msg["header"], None)
        return True
    elif msg_type in ["comm_open", "comm_msg", "comm_close"]:
        with busy_idle(msg["header"]):
            getattr(comm_manager, msg_type)(kernel.shell_stream, None, msg)
        return True
    elif msg_type in ["comm_info_request"]:
        content = msg["content"]
        target_name = msg.get("target_name", None)

        comms = {
            k: dict(target_name=v.target_name) for (k, v) in comm_manager.comms.items() if v.target_name == target_name or target_name is None  # type: ignore
        }
        reply_content = dict(comms=comms, status="ok")
        with busy_idle(msg["header"]):
            msg = session.send(
                kernel.shell_stream,
                "comm_info_reply",
                reply_content,
                msg["header"],
                None,
            )
        return True
    elif msg_type == "shutdown_request":
        send_status("dead", parent=msg["header"])
        return False
    else:
        logger.error("Unsupported msg with msg_type %r", msg_type)
        return False


def read_root(path: str, root_path: str = "", render_kwargs={}, use_nbextensions=True, ssg_data=None) -> Optional[str]:
    if settings.ssg.enabled and ssg_data is None:
        # simply return the pre-rendered html
        from solara_enterprise import ssg

        content = ssg.ssg_content(path)
        if content is not None:
            return content

    default_app = app.apps["__default__"]
    routes = default_app.routes
    router = solara.routing.Router(path, routes)
    if not router.possible_match:
        return None
    if use_nbextensions:
        nbextensions, nbextensions_hashes = get_nbextensions()
    else:
        nbextensions, nbextensions_hashes = [], {}

    from markupsafe import Markup

    def resolve_static_path(path: str) -> Path:
        # this solve a similar problem as the starlette and flask endpoints
        # maybe this can be common code for all of them.
        if path.startswith("/static/public/"):
            directories = [default_app.directory.parent / "public"]
            filename = path[len("/static/public/") :]
        elif path.startswith("/static/assets/"):
            directories = [default_app.directory.parent / "assets", solara_static.parent / "assets"]
            filename = path[len("/static/assets/") :]
        elif path.startswith("/static/"):
            directories = [solara_static.parent / "static"]
            filename = path[len("/static/") :]
        else:
            raise ValueError(f"invalid static path: {path}")
        for directory in directories:
            full_path = directory / filename
            if full_path.exists():
                return full_path
        raise ValueError(f"static path not found: {filename} (path={path}), looked in {directories}")

    def include_css(path: str) -> Markup:
        filepath = resolve_static_path(path)
        content, hash = solara.util.get_file_hash(filepath)
        url = f"{root_path}{path}?v={hash}"
        # when < 10k we embed, also when we use a url, it can be relative, which can break the url
        embed = len(content) < 1024 * 10 and b"url" not in content
        # Always embed the jupyterlab theme CSS to make theme switching possible (see solara.html.j2 template)
        # TODO: Prevent browser from caching the theme CSS files
        if path.endswith("theme-dark.css") or path.endswith("theme-light.css"):
            content_utf8 = content.decode("utf-8")
            code = content_utf8
        elif embed:
            content_utf8 = content.decode("utf-8")
            code = f"<style>/*\npath={path}\n*/\n{content_utf8}</style>"
        else:
            code = f'<link rel="stylesheet" type="text/css" href="{url}">'
        return Markup(code)

    def include_js(path: str, module=False) -> Markup:
        filepath = resolve_static_path(path)
        content, hash = solara.util.get_file_hash(filepath)
        content_utf8 = content.decode("utf-8")
        url = f"{root_path}{path}?v={hash}"
        # when < 10k we embed, but if we use currentScript, it can break things
        embed = len(content) < 1024 * 10 and b"currentScript" not in content
        if embed:
            if module:
                code = f'<script type="module">/*\npath={path}\n*/{content_utf8}</script>'
            else:
                code = f"<script>/*\npath={path}\n*/{content_utf8}</script>"
        else:
            if module:
                code = f'<script type="module" src="{url}"></script>'
            else:
                code = f'<script src="{url}"></script>'
        return Markup(code)

    resources = {
        "theme": "light",
        "nbextensions": nbextensions,
        "nbextensions_hashes": nbextensions_hashes,
        "include_css": include_css,
        "include_js": include_js,
    }
    template: jinja2.Template = get_jinja_env(app_name="__default__").get_template(template_name)
    pre_rendered_html = ""
    pre_rendered_css = ""
    pre_rendered_metas = ""
    title = "Solara ☀️"
    if ssg_data is not None:
        pre_rendered_html = ssg_data["html"]
        pre_rendered_css = "\n".join(ssg_data["styles"])
        pre_rendered_metas = "\n    ".join(ssg_data["metas"])
        title = ssg_data["title"]

    if solara.settings.assets.proxy:
        # solara acts as a proxy
        cdn = f"{root_path}/_solara/cdn"
    else:
        cdn = solara.settings.assets.cdn

    render_settings = {
        "title": title,
        "path": path,
        "root_path": root_path,
        "resources": resources,
        "theme": settings.theme.dict(),
        "production": settings.main.mode == "production",
        "pre_rendered_html": pre_rendered_html,
        "pre_rendered_css": pre_rendered_css,
        "pre_rendered_metas": pre_rendered_metas,
        "assets": settings.assets.dict(),
        "cdn": cdn,
        "ipywidget_major_version": ipywidgets_major,
        "solara_version": solara.__version__,
        "platform": settings.main.platform,
        "vue3": vue3,
        "perform_check": settings.main.mode != "production" and solara.checks.should_perform_solara_check(),
        **render_kwargs,
    }
    response = template.render(**render_settings)
    return response


def find_prefixed_directory(path):
    prefixes = [sys.prefix, os.path.expanduser("~/.local")]
    for prefix in prefixes:
        directory = f"{prefix}{path}"
        if Path(directory).exists():
            return directory
    else:
        raise RuntimeError(f"{path} not found at prefixes: {prefixes}")


@solara.memoize(storage=cache_memory)
def get_nbextensions_directories() -> List[Path]:
    from jupyter_core.paths import jupyter_path

    all_nb_directories = [Path(k) for k in jupyter_path("nbextensions")]
    # FIXME: remove IPython nbextensions path after a migration period
    try:
        from IPython.paths import get_ipython_dir
    except ImportError:
        pass
    else:
        all_nb_directories.append(Path(get_ipython_dir()) / "nbextensions")
    return [Path(k) for k in all_nb_directories]


@solara.memoize(storage=cache_memory)
def get_nbextensions() -> Tuple[List[str], Dict[str, Optional[str]]]:
    from jupyter_core.paths import jupyter_config_path

    paths = [Path(p) / "nbconfig" for p in jupyter_config_path()]

    load_extensions = jupytertools.get_config(paths, "notebook")["load_extensions"]

    def exists(name):
        for directory in nbextensions_directories:
            if (directory / (name + ".js")).exists():
                return True
        logger.info(f"nbextension {name} not found")
        return False

    def hash_extension(name):
        if sys.version_info[:2] < (3, 9):
            # usedforsecurity is only available in Python 3.9+
            h = hashlib.new("md5")
        else:
            h = hashlib.new("md5", usedforsecurity=False)  # type: ignore

        for directory in nbextensions_directories:
            if (directory / (name + ".js")).exists():
                for file in directory.glob("**/*.*"):
                    data = file.read_bytes()
                    h.update(data)

        return h.hexdigest()

    nbextensions: List[str] = [name for name, enabled in load_extensions.items() if enabled and (name not in nbextensions_ignorelist) and exists(name)]
    nbextensions_hashes = {name: hash_extension(name) for name in nbextensions}
    return nbextensions, nbextensions_hashes


nbextensions_directories = get_nbextensions_directories()
solara_static = Path(__file__).parent / "static"



# # tests all pages in the api docs
# from pathlib import Path

# import playwright.sync_api

# import solara

# HERE = Path(__file__).parent


# md = """

# # Solara markdown

# ```solara
# import solara


# @solara.component
# def ClickButton():
#     clicks, set_clicks = solara.use_state(0)
#     def on_click():
#         set_clicks(clicks + 1)
#         print("clicks", clicks)

#     return solara.Button(label=f"Clicked: {clicks}", on_click=on_click)


# app = ClickButton()
# ```

# """


# @solara.component
# def MarkdownApp():
#     return solara.Markdown(md, unsafe_solara_execute=True)


# @solara.component
# def MarkdownAppOff():
#     return solara.Markdown(md, unsafe_solara_execute=False)


# @solara.component
# def MarkdownItApp():
#     return solara.MarkdownIt(md, unsafe_solara_execute=True)


# @solara.component
# def MarkdownItAppOff():
#     return solara.MarkdownIt(md, unsafe_solara_execute=False)


# def test_markdown(page_session: playwright.sync_api.Page, solara_server, solara_app, extra_include_path):
#     with extra_include_path(HERE), solara_app("markdown_test:MarkdownApp"):
#         page_session.goto(solara_server.base_url)
#         page_session.locator("text=Clicked: 0").wait_for()
#         # TODO: flakey test
#         page_session.locator("text=Clicked: 0").click()
#         # page_session.locator("text=Clicked: 1").wait_for()


# def test_markdown_no_execute(page_session: playwright.sync_api.Page, solara_server, solara_app, extra_include_path):
#     with extra_include_path(HERE), solara_app("markdown_test:MarkdownAppOff"):
#         page_session.goto(solara_server.base_url)
#         page_session.locator("text=Solara execution is not enabled").wait_for()


# def test_markdown_it(page_session: playwright.sync_api.Page, solara_server, solara_app, extra_include_path):
#     with extra_include_path(HERE), solara_app("markdown_test:MarkdownItApp"):
#         page_session.goto(solara_server.base_url)
#         page_session.locator("text=Clicked: 0").wait_for()
#         page_session.locator("text=Clicked: 0").click()
#         # TODO: flakey test
#         # page_session.locator("text=Clicked: 1").wait_for()


# def test_markdown_it_no_execute(page_session: playwright.sync_api.Page, solara_server, solara_app, extra_include_path):
#     with extra_include_path(HERE), solara_app("markdown_test:MarkdownItAppOff"):
#         page_session.goto(solara_server.base_url)
#         page_session.locator("text=Solara execution is not enabled").wait_for()



"""
    # Chatbot

    A way to create a chatbot using OpenAI's GPT-4 API, utilizing their new API, and the streaming feature.
"""


import os
from typing import List

from openai import OpenAI
from typing_extensions import TypedDict

import solara
import solara.lab


class MessageDict(TypedDict):
    role: str
    content: str


if os.getenv("OPENAI_API_KEY") is None and "OPENAI_API_KEY" not in os.environ:
    openai = None
else:
    openai = OpenAI()
    openai.api_key = os.getenv("OPENAI_API_KEY")  # type: ignore

messages: solara.Reactive[List[MessageDict]] = solara.reactive([])


def no_api_key_message():
    messages.value = [
        {
            "role": "assistant",
            "content": "No OpenAI API key found. Please set your OpenAI API key in the environment variable `OPENAI_API_KEY`.",
        },
    ]


def add_chunk_to_ai_message(chunk: str):
    messages.value = [
        *messages.value[:-1],
        {
            "role": "assistant",
            "content": messages.value[-1]["content"] + chunk,
        },
    ]


@solara.component
def Page():
    user_message_count = len([m for m in messages.value if m["role"] == "user"])

    def send(message):
        messages.value = [
            *messages.value,
            {"role": "user", "content": message},
        ]

    def call_openai():
        if user_message_count == 0:
            return
        if openai is None:
            no_api_key_message()
            return
        response = openai.chat.completions.create(
            model="gpt-4-1106-preview",
            messages=messages.value,  # type: ignore
            stream=True,
        )
        messages.value = [*messages.value, {"role": "assistant", "content": ""}]
        for chunk in response:
            if chunk.choices[0].finish_reason == "stop":  # type: ignore
                return
            add_chunk_to_ai_message(chunk.choices[0].delta.content)  # type: ignore

    task = solara.lab.use_task(call_openai, dependencies=[user_message_count])  # type: ignore

    with solara.Column(
        style={"width": "700px", "height": "50vh"},
    ):
        with solara.lab.ChatBox():
            for item in messages.value:
                with solara.lab.ChatMessage(
                    user=item["role"] == "user",
                    avatar=False,
                    name="ChatGPT" if item["role"] == "assistant" else "User",
                    color="rgba(0,0,0, 0.06)" if item["role"] == "assistant" else "#ff991f",
                    avatar_background_color="primary" if item["role"] == "assistant" else None,
                    border_radius="20px",
                ):
                    solara.Markdown(item["content"])
        if task.pending:
            solara.Text("I'm thinking...", style={"font-size": "1rem", "padding-left": "20px"})
        solara.lab.ChatInput(send_callback=send, disabled=task.pending)



import pprint

import reacton.ipyvuetify as v

try:
    from solara_enterprise import auth
except ImportError:
    auth = None
import solara as sl


@sl.component
def UserCard():
    user = auth.user.value
    if user:
        user_info = user.get("userinfo")
        if user_info:
            # based on https://v2.vuetifyjs.com/en/components/cards/#props
            with v.Card(width="400px"):
                with v.ListItem(three_line=True):
                    with v.ListItemContent():
                        sl.Div("Logged in", class_="text-overline mb-4")
                        v.ListItemTitle(children=[user_info["email"]])
                        v.ListItemSubtitle(children=["You are now logged in, log out via the app bar, or the button below"])
                    with v.ListItemAvatar():
                        auth.Avatar()

                with v.CardActions():
                    sl.Button("logout", icon_name="mdi-logout", href=auth.get_logout_url(), text=True)
        else:
            sl.Error("No user info")
    else:
        sl.Error("No user")


@sl.component
def Page():
    sl.Title("Login demo using OAuth")
    with sl.AppBar():
        if auth.user.value:
            auth.AvatarMenu()
        else:
            sl.Button(icon_name="mdi-login", href=auth.get_login_url(), icon=True)

    with sl.Column():
        if auth.user.value:
            UserCard()
            with sl.Details("Raw data"):
                sl.Markdown(
                    """
                    ### Raw user data

                    *Note: do not share this data with anyone, it contains sensitive information.*

                    This is the raw user data from the auth provider.

                    We use the `picture` field to display an avatar in the [AppBar](/api/appbar).
                """
                )
                sl.Preformatted(pprint.pformat(auth.user.value))
        else:
            sl.Markdown(
                """
            ### Login demo

            This is a demo of a login system using OAuth. You can login with your google account, github account or with a username and password.
            We are using [Auth0](https://auth0.com/) as an OAuth provider.

            """
            )
            with sl.Row():
                sl.Button("login", icon_name="mdi-login", href=auth.get_login_url())


if auth is None:
    del Page



import solara


@solara.component
def Page():
    with solara.Link("/showcase"):
        solara.Text("« Back to Showcases")
    with solara.ColumnsResponsive(12, medium=6):
        solara.Markdown(
            """
        # Solara.dev

        The Solara.dev website is built using Solara itself, showcasing the library's features and capabilities.
        All documentation and example are on this website.
        """
        )

    with solara.ColumnsResponsive(12, medium=6):
        solara.Markdown(
            """
        ## Desktop and mobile

        Solara.dev is designed to be responsive, working seamlessly on both desktop and mobile devices.
        On smaller screens, the top navigation bar is replaced with a drawer for easy navigation.
        The [`ColumnsResponsive`](/api/columns_responsive) component can be used to create responsive content layouts.

        """
        )
        solara.Image("https://dxhl76zpt6fap.cloudfront.net/public/showcase/solara/home-mobile.png", width="100%", classes=["pt-12"])

    with solara.ColumnsResponsive(12, medium=6):
        solara.Image("https://dxhl76zpt6fap.cloudfront.net/public/showcase/solara/docs.png", width="100%")
        solara.Markdown(
            """
## Multi-page

The website is composed of multiple pages, illustrating how users can navigate between the home page, documentation, and tutorial sections.

        """
        )

    with solara.ColumnsResponsive(12, medium=6):
        solara.Markdown(
            """
            ## Markdown support

            Solara supports the use of Markdown for creating dynamic and static content.
            This makes it possible to combine Python-generated pages with static Markdown pages in a single website.
            Hot reloading ensures that the page is automatically updated whenever changes are made to the Markdown files.


        """
        )
        solara.Image("https://dxhl76zpt6fap.cloudfront.net/public/showcase/solara/vscode-markdown.png", width="100%")



"""# Image annotation with Solara

This example displays how to annotate images with different drawing tools in plotly figures. Use the canvas
below to draw shapes and visualize the canvas callback.


Check [plotly docs](https://dash.plotly.com/annotations) for more information about image annotation.
"""
import json

import plotly.graph_objects as go

import solara

title = "Plotly Image Annotator"
shapes = solara.reactive(None)


class CustomEncoder(json.JSONEncoder):
    """
    Custom JSON encoder for Plotly objects.

    Plotly may return objects that the standard JSON encoder can't handle. This
    encoder converts such objects to str, allowing serialization by json.dumps
    """

    def default(self, o):
        if isinstance(o, object):
            return str(o)
        return super().default(o)


@solara.component
def Page():
    def on_relayout(data):
        if data is None:
            return

        relayout_data = data["relayout_data"]

        if "shapes" in relayout_data:
            shapes.value = relayout_data["shapes"]

    fig = go.FigureWidget(
        layout=go.Layout(
            showlegend=False,
            autosize=False,
            width=600,
            height=600,
            dragmode="drawrect",
            modebar={
                "add": [
                    "drawclosedpath",
                    "drawcircle",
                    "drawrect",
                    "eraseshape",
                ]
            },
        )
    )

    solara.FigurePlotly(fig, on_relayout=on_relayout)
    if not shapes.value:
        solara.Markdown("## Draw on the canvas")
    else:
        solara.Markdown("## Data returned by drawing")
        formatted_shapes = str(json.dumps(shapes.value, indent=2, cls=CustomEncoder))
        solara.Preformatted(formatted_shapes)



import solara


@solara.component
def Page():
    with solara.Link("/showcase"):
        solara.Text("« Back to Showcases")
    with solara.ColumnsResponsive(12, medium=6):
        solara.Markdown(
            """
        # TESSA by Planeto

        [Planeto](https://planeto-energy.ch/) developed a tool called [TESSA](https://planeto-energy.ch/solution/) for district heating & cooling planning.

        TESSA was prototyped in the Jupyter notebook using ipywidgets. Using solara, they are able to bring TESSA into production using the
        same technology stack.
        """
        )
        solara.Image("https://dxhl76zpt6fap.cloudfront.net/public/showcase/tessa/thumbnail.png", width="100%", classes=["pt-12"])



"""# Info

Solara has 4 types of alerts:

 * [Success](/api/success)
 * Info (this page)
 * [Warning](/api/warning)
 * [Error](/api/error)



"""

import solara
from solara.website.utils import apidoc


@solara.component
def Page():
    icon, set_icon = solara.use_state(True)
    dense, set_dense = solara.use_state(False)
    outlined, set_outlined = solara.use_state(True)
    text, set_text = solara.use_state(True)
    with solara.VBox() as main:
        with solara.GridFixed(4):
            solara.Checkbox(label="Use icon", value=icon, on_value=set_icon)
            solara.Checkbox(label="Show dense", value=dense, on_value=set_dense)
            solara.Checkbox(label="Show as text", value=text, on_value=set_text)
            solara.Checkbox(label="Show outlined", value=outlined, on_value=set_outlined)
        solara.Info(
            f"This is solara.Info(label='...', text={text}, dense={dense}, outlined={outlined}, icon={icon})",
            text=text,
            dense=dense,
            outlined=outlined,
            icon=icon,
        )
    return main


__doc__ += apidoc(solara.Info.f)  # type: ignore



"""# Input fields

This page contains all the input fields available in Solara.

# InputText
"""
import solara
from solara.website.utils import apidoc

from . import NoPage

Page = NoPage


__doc__ += apidoc(solara.InputText.f)  # type: ignore
__doc__ += "# InputFloat"
__doc__ += apidoc(solara.InputFloat.f)  # type: ignore
__doc__ += "# InputInt"
__doc__ += apidoc(solara.InputInt.f)  # type: ignore



import dataclasses

import solara
from solara.alias import rw


@dataclasses.dataclass
class Clicks:
    value: int


@solara.component
def ButtonClick():
    clicks, set_clicks = solara.use_state(Clicks(0))
    return rw.Button(description=f"Clicked {clicks.value} times", on_click=lambda: set_clicks(Clicks(clicks.value + 1)))



import itertools
from typing import Dict, List, Union

import reacton.ipyvuetify as rv

import solara
from solara.util import _combine_classes


def cycle(value):
    if value is None:
        return itertools.cycle([None])
    elif isinstance(value, int):
        return itertools.cycle([value])
    elif isinstance(value, (list, tuple)):
        return itertools.cycle(value)
    else:
        raise ValueError(f"Invalid value for columns: {value}, should be None, int, or list/tuple.")


@solara.component
def Columns(
    widths: List[Union[float, int]] = [1],
    wrap=True,
    gutters=True,
    gutters_dense=False,
    children=[],
    classes: List[str] = [],
    style: Union[str, Dict[str, str], None] = None,
):
    """Lays out children in columns, with relative widths specified as a list of floats or ints.

    Widths are relative to each other, so [1, 2, 1] will result in 1/4, 1/2, 1/4 width columns.
    Columns with a width of 0 will take up the minimal amount of space.

    If there are more children than widths, the width list will be cycled.

    ```python
    with solara.Columns([1, 2, 1]):
        solara.Text("I am on the left")
        with solara.Card("Middle"):
            ...
        with solara.Column():
            ...
    ```

    When three children are added to this component, they will be laid out in three columns,
    with the first and last column taking up 1/4 of the width, and the middle column taking up 1/2 of the width.

    ```solara
    import solara

    @solara.component
    def Page():
        with solara.Columns([0, 1, 2]):
            solara.Text("I am as small as possible")
            solara.Select("I stretch", values=["a", "b", "c"], value="a")
            solara.Select("I stretch twice the amount", values=["a", "b", "c"], value="a")
    ```


    # Arguments

     * `widths`: List of floats or ints, specifying the relative widths of the columns.
     * `wrap`: Whether to wrap the columns to the next row if there is not enough space available. This only happens when using widths of 0.
     * `gutters`: Whether to add gutters between the columns.
     * `gutters_dense`: Make gutters smaller.
     * `children`: List of children to be laid out in columns.
     * `style`: CSS style to apply to the top level element.
     * `classes`: List of CSS classes to be applied to the top level element.

    """
    class_ = _combine_classes([*(["flex-nowrap"] if not wrap else []), *classes])
    style_flat = solara.util._flatten_style(style)
    with rv.Row(class_=class_, no_gutters=not gutters, dense=gutters_dense, style_=style_flat) as main:
        for child, width in zip(children, cycle(widths)):
            # we add height: 100% because this will trigger a chain of height set if it is set on the parent
            # via the style. If we do not set the height, it will have no effect. Furthermore, we only have
            # a single child, so this cannot interfere with other siblings.
            with rv.Col(children=[child], style_=f"height: 100%; flex-grow: {width}; overflow: auto" if width != 0 else "flex-grow: 0"):
                pass
    return main


@solara.component
def ColumnsResponsive(
    default=None,
    small=None,
    medium=None,
    large=None,
    xlarge=None,
    children=[],
    wrap=True,
    gutters=True,
    gutters_dense=False,
    classes: List[str] = [],
    style: Union[str, Dict[str, str], None] = None,
):
    """Lay our children in columns, on a 12 point grid system that is responsive to screen size.

    If a single number is specified, or less values than children, the values will be cycled.
    The total width of this system is 12, so if you want to have 3 columns, each taking up 4 points, you would specify [4, 4, 4] or 4.


    ```python
    with ColumnsResponsive([4, 4, 4]):
        ...
    with ColumnsResponsive(4):  # same effect
        ...
    ```

    If you want the first column to take up 4 points, and the second column to take up the remaining 8 points, you would specify [4, 8].

    ```python
    with ColumnsResponsive([4, 8]):
        ...
    ```

    If you want your columns to be full width on large screen, and next to each other on larger screens.

    ```python
    with ColumnsResponsive(12, large=[4, 8]):
        ...
    ```

    # Arguments

     * default: Width of column for >= 0 px.
     * small: Width of column for >= 600 px.
     * medium: Width of column >= 960 px.
     * large: Width of column for >= 1264 px.
     * xlarge: Width of column for >= 1904 px.

    """

    def cycle(value):
        if value is None:
            return itertools.cycle([None])
        elif isinstance(value, int):
            return itertools.cycle([value])
        elif isinstance(value, (list, tuple)):
            return itertools.cycle(value)
        else:
            raise ValueError(f"Invalid value for columns: {value}, should be None, int, or list/tuple.")

    class_ = _combine_classes([*(["flex-nowrap"] if not wrap else []), *classes])
    style_flat = solara.util._flatten_style(style)
    with rv.Row(class_=class_ if not wrap else "", style_=style_flat, no_gutters=not gutters, dense=gutters_dense) as main:
        for child, xsmall, small, medium, large, xlarge in zip(children, cycle(default), cycle(small), cycle(medium), cycle(large), cycle(xlarge)):
            with rv.Col(
                cols=xsmall,
                sm=small,
                md=medium,
                lg=large,
                xl=xlarge,
                children=[child],
            ):
                pass
    return main



"""
# ConfirmationDialog

"""

from typing import Union

import solara
from solara.lab.components.confirmation_dialog import ConfirmationDialog
from solara.website.utils import apidoc

title = "ConfirmationDialog"
users = solara.reactive("Alice Bob Cindy Dirk Eve Fred".split())
user_to_be_deleted: solara.Reactive[Union[str, None]] = solara.reactive(None)


def ask_to_delete_user(user):
    user_to_be_deleted.value = user


def clear_user_to_be_deleted():
    user_to_be_deleted.value = None


def delete_user():
    users.set([u for u in users.value if u != user_to_be_deleted.value])
    clear_user_to_be_deleted()


@solara.component
def Page():
    """Create a list of users with a button to delete them.

    A confirmation dialog will pop up first before deletion."""
    solara.Markdown("#### Users:")
    with solara.Column(style={"max-width": "300px"}):
        for user in users.value:
            with solara.Row(style={"align-items": "center"}):
                solara.Text(user)
                solara.v.Spacer()
                solara.Button(icon_name="mdi-delete", on_click=lambda user=user: ask_to_delete_user(user), icon=True)
        if not users.value:
            solara.Text("(no users left)")

    with ConfirmationDialog(
        user_to_be_deleted.value is not None,
        on_ok=delete_user,
        on_close=clear_user_to_be_deleted,
        ok="Ok, Delete",
        title="Delete user",
    ):
        solara.Markdown(f"Are you sure you want to delete user **{user_to_be_deleted.value}**?")


__doc__ += apidoc(solara.lab.components.confirmation_dialog.ConfirmationDialog.f)  # type: ignore



import ipywidgets as widgets
from IPython.display import display

import solara.components.file_download

# test if normal widget code works with no app context


def test_create_widget(no_kernel_context):
    button = widgets.Button(description="Click me")
    button.layout.close()
    button.close()


def test_vue_template(no_kernel_context):
    widget = solara.components.file_download.FileDownloadWidget()
    widget.close()


def test_display(no_kernel_context):
    display("test")



import solara

from . import SharedComponent

github_url = solara.util.github_url(__file__)


@solara.component
def Page():
    component_state = solara.use_reactive(0)
    solara.Title("Page 2 set with title")
    with solara.Card("Page 2"):
        solara.Markdown("Page 2 is even better, even though it has no sidebar.")
        solara.Button(label="View source", icon_name="mdi-github-circle", attributes={"href": github_url, "target": "_blank"}, text=True, outlined=True)
        SharedComponent()

        with solara.Card("Shown only on page2", style={"max-width": "500px"}, margin=0, classes=["my-2"]):
            solara.Markdown(
                f"""
                The lifetime of `component_state` if bound to this page. If you navigate away from it,
                the state will be reset.

                It creates the `component_state` [using the use_reactive hook](https://solara.dev/api/use_reactive)
                so that the state is bound to the component.

                See also [about state management](https://solara.dev/docs/fundamentals/state-management) for more information.


                component_state: {component_state.value}
                """
            )
            solara.Button(
                label="Increment component_state", icon_name="mdi-plus", on_click=lambda: component_state.set(component_state.value + 1), outlined=True
            )
